from enum import IntFlag
from typing import Any, Dict, List, Optional, TypeVar, Union

from mstrio.api import browsing, objects
from mstrio.connection import Connection
from mstrio.utils.entity import Entity, ObjectTypes
from mstrio.utils.helper import fetch_objects


class ShortcutInfoFlags(IntFlag):
    DssDossierShortcutInfoBookmark = 0b10
    DssDossierShortcutInfoTOC = 0b01
    DssDossierShortcutInfoDefault = 0b00


class Shortcut(Entity):
    _OBJECT_TYPE = ObjectTypes.SHORTCUT_TYPE
    _API_GETTERS = {
        ('name', 'id', 'project_id', 'owned_by_current_user', 'target', 'encode_html_content',
         'current_page_key', 'shared_time', 'last_viewed_time', 'last_modified_time', 'stid',
         'current_bookmark', 'prompted', 'datasets_cache_info_hash', 'shortcut_info_flag',
         'dossier_version_hash'): browsing.get_shortcut,
        ('abbreviation', 'type', 'subtype', 'ext_type', 'date_created', 'date_modified', 'version',
         'owner', 'icon_path', 'view_media', 'ancestors', 'certified_info', 'acg', 'acl',
         'target_info'): objects.get_object_info
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'shortcut_info_flag': ShortcutInfoFlags,
    }

    def __init__(
            self, connection: Connection, id: str, project_id: str,
            shortcut_info_flag: ShortcutInfoFlags = ShortcutInfoFlags.DssDossierShortcutInfoTOC):
        """Initialize the Shortcut object and populate it with I-Server data.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            id: Shortcut ID
            project_id: ID of the project that the shortcut is in
            shortcut_info_flag: flag indicating what information about shortcut
                should be loaded
        """
        if id is None or project_id is None:
            raise AttributeError(
                "Please specify 'id' and 'project_id' parameters in the constructor.")
        else:
            super().__init__(connection=connection, object_id=id, project_id=project_id,
                             shortcut_info_flag=shortcut_info_flag.value)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._owned_by_current_user = kwargs.get('owned_by_current_user')
        self._target = kwargs.get('target')
        self._encode_html_content = kwargs.get('encode_html_content')
        self._shortcut_info_flag = ShortcutInfoFlags(
            kwargs["shortcut_info_flag"]) if kwargs.get("shortcut_info_flag") else None
        self._current_page_key = kwargs.get('current_page_key')
        self._shared_time = kwargs.get('shared_time')
        self._last_viewed_time = kwargs.get('last_viewed_time')
        self._last_modified_time = kwargs.get('last_modified_time')
        self._stid = kwargs.get('stid')
        self._current_bookmark = kwargs.get('current_bookmark')
        self._prompted = kwargs.get('prompted')
        self._datasets_cache_info_hash = kwargs.get('datasets_cache_info_hash')
        self._dossier_version_hash = kwargs.get('dossier_version_hash')

    T = TypeVar('T')

    @classmethod
    def from_dict(cls: T, source: Dict[str, Any], connection: Connection) -> Union[T, dict]:
        """Instantiate a Shortcut from a dict source.
           If `projectId` not present in source, return dict."""
        # Without projectId the fetching functionality will fail
        if source.get("project_id") or source.get("projectId"):
            return super(Entity, cls).from_dict(source=source, connection=connection)
        return source

    @property
    def owned_by_current_user(self):
        return self._owned_by_current_user

    @property
    def target(self):
        return self._target

    @property
    def encode_html_content(self):
        return self._encode_html_content

    @property
    def shortcut_info_flag(self):
        return self._shortcut_info_flag

    @property
    def current_page_key(self):
        return self._current_page_key

    @property
    def shared_time(self):
        return self._shared_time

    @property
    def last_viewed_time(self):
        return self._last_viewed_time

    @property
    def last_modified_time(self):
        return self._last_modified_time

    @property
    def stid(self):
        return self._stid

    @property
    def current_bookmark(self):
        return self._current_bookmark

    @property
    def prompted(self):
        return self._prompted

    @property
    def datasets_cache_info_hash(self):
        return self._datasets_cache_info_hash

    @property
    def dossier_version_hash(self):
        return self._dossier_version_hash


def get_shortcuts(
        connection: Connection, project_id: str, shortcut_ids: List[str],
        shortcut_info_flag: ShortcutInfoFlags = ShortcutInfoFlags.DssDossierShortcutInfoDefault,
        to_dictionary: bool = False, limit: Optional[int] = None,
        **filters) -> Union[List[dict], List[Shortcut]]:
    """Retrieve information about specific published shortcuts
    in specific project.

    Args:
        shortcut_ids: ids of target shortcuts
        project_id: id of project that the shortcuts are in
        shortcut_info_flag: a single ShortcutInfoFlags that describes what
          exact info are to be fetched
        to_dictionary: parameter describing output format
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
    Return:
        list of dictionaries or Shortcut objects,
          depending on `to_dictionary` parameter
    """

    shortcuts = fetch_objects(
        connection=connection,
        api=browsing.get_shortcuts,
        dict_unpack_value="shortcuts",
        limit=limit,
        filters=filters,
        body=[{
            "projectId": project_id,
            "shortcutIds": shortcut_ids
        }],
        shortcut_info_flag=shortcut_info_flag.value,
    )

    if to_dictionary:
        return shortcuts
    else:
        return [Shortcut.from_dict(source=short, connection=connection) for short in shortcuts]



"""This is the demo script to show how administrator can manage documents and
dossiers in users' libraries.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.project_objects import Document, list_documents, Dossier, list_dossiers
from mstrio.users_and_groups import User, list_users, UserGroup, list_user_groups
from mstrio.connection import Connection

base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
conn = Connection(base_url, username, password, project_name='MicroStrategy Library',
                  login_mode=1)

# list all dossiers and documents within a project to which we have connection
dossiers = list_dossiers(connection=conn)
docs = list_documents(connection=conn)

# get document and dossier from by name or id and publish them to a library of
# an authenticated user
doc = Document(connection=conn, name='Some Simple Document')
doss = Dossier(connection=conn, name='Some Simple Dossier')
doc.publish()
doss.publish()

# list all users and get 2 of them
users = list_users(connection=conn)
user_1 = User(connection=conn, id='1234234534564567567867897890ABCD')
user_2 = User(connection=conn, id='9876876576546543543243213210DCBA')

# share one dossier/document to a given user(s) by passing user object(s)
# or id(s)
doss.publish(recipients=user_1)
doss.publish(recipients=['1234234534564567567867897890ABCD', '9876876576546543543243213210DCBA'])
# analogously we can take away dossier(s)/document(s) from the library of the
# given user(s)
doss.unpublish(recipients=[user_1, user_2])
doss.unpublish(recipients='1234234534564567567867897890ABCD')

# list all user groups and get one of them
user_groups_ = list_user_groups(connection=conn)
user_group_ = UserGroup(connection=conn, name='Data Scientists')

# add users to given user group
user_group_.add_users(users=[user_1, user_2])

# get documents with given ids to give to the user group and users which belong
# to it
docs_to_publish = list_documents(
    connection=conn,
    id=[
        '12340987234598763456876545677654', '98761234876523457654345665434567',
        '654356785432678943217890ABCDDCBA'
    ],
)
for doc in docs_to_publish:
    doc.publish(recipients=user_group_)



from typing import TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@ErrorHandler(err_msg='Error creating a new changeset.')
def create_changeset(connection: "Connection", project_id: str = None, schema_edit: bool = False,
                     error_msg: str = None):
    """Create a new changeset for modelling manipulations."""
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id
    return connection.session.post(
        url=f"{connection.base_url}/api/model/changesets",
        headers={'X-MSTR-ProjectID': project_id},
        params={'schemaEdit': str(schema_edit).lower()}
    )


@ErrorHandler(err_msg='Error committing changeset {id} changes to the metadata.')
def commit_changeset_changes(connection: "Connection", id: str, error_msg: str = None,
                             throw_error=True):
    """Commit the changeset changes to metadata."""
    return connection.session.post(
        url=f"{connection.base_url}/api/model/changesets/{id}/commit",
        headers={'X-MSTR-MSChanget': id}, params={'changesetId': id}
    )


@ErrorHandler(err_msg='Error deleting the changeset with ID {id}')
def delete_changeset(connection: "Connection", id: str, error_msg: str = None):
    """Delete the changeset."""
    return connection.session.delete(
        url=f"{connection.base_url}/api/model/changesets/{id}",
        headers={'X-MSTR-MSChanget': id},
        params={'changesetId': id}
    )



# TODO: fix circular import
# flake8: noqa
# from .privilege_mode import PrivilegeMode
# from .privilege import Privilege, PrivilegeList
# from .security_role import list_security_roles, SecurityRole



# flake8: noqa
from .dbms import Dbms, list_available_dbms
from .database_connections import DatabaseConnections
from .datasource_connection import (CharEncoding, DatasourceConnection, DriverType, ExecutionMode,
                                    list_datasource_connections)
from .datasource_instance import (DatasourceInstance, DatasourceType, list_datasource_instances,
                                  DatasourceType)
from .datasource_login import DatasourceLogin, list_datasource_logins
from .datasource_map import DatasourceMap, list_datasource_mappings, Locale



from typing import List, Union

from mstrio.connection import Connection
from mstrio.utils import helper

from ..schedule import Schedule
from .base_subscription import Subscription
from .content import Content


class EmailSubscription(Subscription):
    """Class representation of MicroStrategy Email Subscription object."""

    def __init__(self, connection: Connection, subscription_id: str = None, project_id: str = None,
                 project_name: str = None, application_id: str = None,
                 application_name: str = None):
        """Initialize EmailSubscription object, populates it with I-Server data
        if subscription_id is passed.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is
        omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
            subscription_id: ID of the subscription to be initialized
            project_id: Project ID
            project_name: Project name
            application_id: deprecated. Use project_id instead.
            application_name: deprecated. Use project_name instead.
        """
        if application_id or application_name:
            helper.deprecation_warning(
                '`application_id` and `application_name`',
                '`project_id` and `project_name`',
                '11.3.4.101',  # NOSONAR
                False)
            project_id = project_id or application_id
            project_name = project_name or application_name
        if subscription_id:
            super().__init__(connection, subscription_id, project_id, project_name)

    @classmethod
    def create(cls, connection: Connection, name: str, recipients: Union[List[str], List[dict]],
               project_id: str = None, schedules: Union[str, List[str], Schedule,
                                                        List[Schedule]] = None,
               project_name: str = None, allow_delivery_changes: bool = None,
               allow_personalization_changes: bool = None, allow_unsubscribe: bool = True,
               send_now: bool = None, owner_id: str = None, contents: Content = None,
               delivery_expiration_date: str = None, contact_security: bool = None,
               email_subject: str = None, email_message: str = None, filename: str = None,
               compress: bool = False, space_delimiter: str = None,
               email_send_content_as: str = 'data', overwrite_older_version: bool = False,
               zip_filename: str = None, zip_password_protect: bool = None,
               zip_password: str = None, application_id: str = None, application_name: str = None):
        """Creates a new email subscription.

        Args:
            connection(Connection): a MicroStrategy connection object
            name(str): name of the subscription,
            project_id(str): project ID,
            project_name(str): project name,
            application_id: deprecated. Use project_id instead.
            application_name: deprecated. Use project_name instead.
            allow_delivery_changes(bool): whether the recipients can change
                the delivery of the subscription,
            allow_personalization_changes(bool): whether the recipients can
                personalize the subscription,
            allow_unsubscribe(bool): whether the recipients can unsubscribe
                from the subscription,
            send_now(bool): indicates whether to execute the subscription
                immediately,
            owner_id(str): ID of the subscription owner, by default logged in
                user ID,
            schedules (Union[str, List[str], Schedule, List[Schedule]]):
                Schedules IDs or Schedule objects,
            contents(Content): The content settings.
            recipients(Union[List[str], List[dict]]): list of recipients IDs
                or dicts,
            delivery_expiration_date(str): expiration date of the subscription,
                format should be yyyy - MM - dd,
            contact_security(bool): whether to use contact security for each
                contact group member,
            filename(str): the filename that will be delivered when
                the subscription is executed,
            compress(bool): whether to compress the file
            space_delimiter(str): space delimiter,
            email_subject(str): email subject associated with the subscription,
            email_message(str): email body of subscription,
            email_send_content_as(str, enum): [data, data_and_history_list,
                data_and_link_and_history_list, link_and_history_list],
            overwrite_older_version(bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            zip_filename(str): filename of the compressed content,
            zip_password_protect(bool): whether to password protect zip file,
            zip_password(str): optional password for the compressed file
        """
        if application_id or application_name:
            helper.deprecation_warning(
                '`application_id` and `application_name`',
                '`project_id` and `project_name`',
                '11.3.4.101',  # NOSONAR
                False)
            project_id = project_id or application_id
            project_name = project_name or application_name

        return super()._Subscription__create(
            connection=connection,
            name=name,
            project_id=project_id,
            project_name=project_name,
            allow_delivery_changes=allow_delivery_changes,
            allow_personalization_changes=allow_personalization_changes,
            allow_unsubscribe=allow_unsubscribe,
            send_now=send_now,
            owner_id=owner_id,
            schedules=schedules,
            contents=contents,
            recipients=recipients,
            delivery_mode='EMAIL',
            delivery_expiration_date=delivery_expiration_date,
            contact_security=contact_security,
            email_subject=email_subject,
            email_message=email_message,
            filename=filename,
            compress=compress,
            space_delimiter=space_delimiter,
            email_send_content_as=email_send_content_as,
            overwrite_older_version=overwrite_older_version,
            zip_filename=zip_filename,
            zip_password_protect=zip_password_protect,
            zip_password=zip_password,
        )



from typing import TYPE_CHECKING, Union, List

from mstrio.project_objects.document import list_documents, Document
from mstrio.project_objects.dossier import list_dossiers
from mstrio.api import library
from mstrio.utils.helper import exception_handler

if TYPE_CHECKING:
    from mstrio.project_objects.dossier import Dossier


class Library:

    def __init__(self, connection):
        # TODO: consider adding Connection.project_selected attr/method
        if connection.project_id is None:
            exception_handler("No project selected, library content will not be loaded.",
                              exception_type=Warning)
        self.connection = connection
        ids = self.__get_library_ids()
        # TODO: consider adding Connection.project_selected attr/method
        if connection.project_id is None:
            self._documents = None
            self._dossiers = None
            self._contents = None
        else:
            self._documents = list_documents(self.connection, id=ids)
            self._dossiers = list_dossiers(self.connection, id=ids)
            self._contents = self._documents + self._dossiers
        self.user_id = connection.user_id

    def __get_library_ids(self):
        response = library.get_library(self.connection)
        body = response.json()
        ids = [doc_body['target']['id'] for doc_body in body if 'target' in doc_body]
        return ids

    @property
    def dossiers(self):
        if self.connection.project_id is not None:
            ids = self.__get_library_ids()
            self._dossiers = list_dossiers(self.connection, id=ids)
        return self._dossiers

    @property
    def documents(self):
        if self.connection.project_id is not None:
            ids = self.__get_library_ids()
            self._documents = list_documents(self.connection, id=ids)
        return self._documents

    @property
    def contents(self):
        if self.connection.project_id is not None:
            self._contents = self.dossiers + self.documents
        return self._contents

    def publish(self, contents: Union[List, "Document", "Dossier", str]):
        """Publishes dossier or document to the authenticated user's library.

        contents: dossiers or documents to be published, can be Dossier/Document
            class object or ID
        """
        if not isinstance(contents, list):
            contents = [contents]
        for doc in contents:
            doc_id = doc.id if isinstance(doc, Document) else doc
            body = {'id': doc_id, 'recipients': [{'id': self.user_id}]}
            library.publish_document(self.connection, body=body)

    def unpublish(self, contents: Union[List, "Document", "Dossier", str]):
        """Publishes dossier or document to the authenticated user's library.

        contents: dossiers or documents to be published, can be Dossier/Document
            class object or ID
        """
        if not isinstance(contents, list):
            contents = [contents]
        for doc in contents:
            doc_id = doc.id if isinstance(doc, Document) else doc
            library.unpublish_document_for_user(
                self.connection,
                user_id=self.user_id,
                document_id=doc_id,
            )



from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, Optional, Union

from requests_futures.sessions import FuturesSession

from mstrio import config
from mstrio.api import monitors
from mstrio.connection import Connection
from mstrio.server.cluster import Cluster
from mstrio.utils import helper


class DatabaseConnections:
    """Browse and manage database connections on the environment.

    Attributes:
        connection: A MicroStrategy connection object
    """

    def __init__(self, connection: Connection):
        """Initialize the `DatabaseConnections` object.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
        """
        self.connection = connection

    def list_connections(self, nodes: Union[str, List[str]] = None, limit: Optional[int] = None,
                         **filters) -> List[Dict[str, Any]]:
        """Get all active database connections. Optionally filter the
         connections by specifying the `filters` keyword arguments.

        Args:
            nodes: Node (server) names on which databases will be disconnected.
            limit: limit the number of elements returned. If `None`, all objects
                are returned.
            **filters: Available filter parameters: ['status',
                'database_instance_name', 'database_instance_id', 'user_name',
                'database_login_name', 'cluster_node', 'id', 'name', 'type']
        """
        if nodes is None:
            all_nodes = Cluster(self.connection).list_nodes(to_dictionary=True)
            nodes = [n['name'] for n in all_nodes if n['status'] == 'running']

        nodes_names = ",".join(nodes) if isinstance(nodes, list) else nodes
        all_databases = helper.fetch_objects(connection=self.connection,
                                             api=monitors.get_database_connections,
                                             nodes_names=nodes_names,
                                             dict_unpack_value="dbConnectionInstances",
                                             limit=limit, filters=filters)
        return all_databases

    def disconnect_database(self, connection_id: str, force: bool = False) -> bool:
        """Disconnect database connections by passing in connection_id.

        Args:
            connection_id: Database Connection Instance Id
            force: if True, no additional prompt will be shown before.
                Default False.
            Returns:
                True for success. False otherwise.
        """
        user_input = 'N'
        if not force:
            user_input = input((f"Are you sure you want to disconnect database connection "
                                f"'with ID:{connection_id}? [Y/N]: "))
        if force or user_input == 'Y':
            response = monitors.delete_database_connection(self.connection, connection_id)
            if response.status_code == 204 and config.verbose:
                print(f"Successfully disconnected database connection instance {connection_id}.")
            return response.ok
        else:
            return False

    def disconnect_all_databases(self, force: bool = False) -> Union[List[dict], None]:
        """Disconnect all database connections.

        Args:
            force: if True, no additional prompt will be shown before
                disconnecting all connections

        Returns:
            - list of statuses of disconnecting all connections with their ids
               and messages from the I-Server
            - in case of error it returns None
        """
        user_input = 'N'
        if not force:
            user_input = input(
                "Are you sure you want to disconnect all database connections? [Y/N]: ")
            if user_input != "Y":
                return None
            else:
                force = True

        connections = self.list_connections()
        threads = helper.get_parallel_number(len(connections))
        with FuturesSession(executor=ThreadPoolExecutor(max_workers=threads),
                            session=self.connection.session) as session:

            futures = (monitors.delete_database_connection_async(session, self.connection,
                                                                 conn["id"])
                       for conn in connections)
            statuses: List[Dict[str, Union[str, int]]] = []
            for f in futures:
                response = f.result()
                statuses.append({
                    'id': response.url.rsplit("/").pop(-1),
                    'status': response.status_code
                })
        return self._prepare_disconnect_by_id_message(statuses=statuses)

    @staticmethod
    def _prepare_disconnect_by_id_message(statuses: List[dict]) -> List[dict]:
        succeeded = []
        failed = []

        for s in statuses:
            status = s['status']
            if status < 400:
                succeeded.append(s['id'])
            else:
                # in most cases there will be status code 403
                failed.append(s['id'])

        if config.verbose:
            if succeeded:
                print("Database connections with ids listed "
                      "below were successfully disconnected:\n\t" + ',\n\t'.join(succeeded))
            if failed:
                print("Database connections with ids listed "
                      "below were not disconnected:\n\t" + ',\n\t'.join(failed))
        return statuses



from enum import Enum
from mstrio.utils.helper import Dictable
from typing import Optional

from mstrio.access_and_security.security_filter import ObjectReference, ObjectInformation

# TODO maybe enhance Qualification of security filter definition
# with given objects


class TokenTextType(Enum):
    END_OF_TEXT = 'end_of_text'
    ERROR = 'error'
    UNKNOWN = 'unknown'
    EMPTY = 'empty'
    CHARACTER = 'character'
    LITERAL = 'literal'
    IDENTIFIER = 'identifier'
    STRING_LITERAL = 'string_literal'
    INTEGER = 'integere'
    FLOAT = 'float'
    BOOLEAN = 'boolean'
    GUID = 'guid'
    OBJECT_REFERENCE = 'object_reference'
    COLUMN_REFERENCEE = 'column_reference'
    OBJECT_AT_FORM = 'object_at_form'
    FUNCTION = 'function'
    KEYWORD = 'keyword'
    OTHER = 'other'


class TokenLevel(Enum):
    CLIENT = 'client'
    LEXED = 'lexed'
    RESOLVED = 'resolved'
    PARSED = 'parsed'


class TokenState(Enum):
    ERROR = 'error'
    INITIAL = 'initial'
    OKAY = 'okay'


class ParserToken(Dictable):
    """Structure that represents a fragment of text that should be handled as
    one entity when parsing.

    The expression parser annotates the text it parses by converting it into an
    array of token objects. These objects are used for several purposes:
        - they allow us to record extra information (generally the identity of
          an object) that is not explicitly shown in the raw text. Thus once we
          have disambiguated a name, it is not necessary to do so again.
        - they permit a GUI to reformat the text (using colour, type-face etc.)
          to make it easier to understand the expression.
        - they allow the parser to report the location of parse errors within
          the raw text.

    Attributes:
        value (string): the raw text represented by this token.
        type (enum): enumeration constant that classifies the text within this
            token
        level (enum): describe the amount of processing performed on this
            parser token
        state (enum): whether token is an error or not
        target (object): if the token represents an object, provide information
            about the object
        attribute_form (object): if the token represents an attribute form in
            the context of an object (say `City@DESC`) then provide information
            about the attribute form.
    """

    def __init__(self, value: str, type: Optional[TokenTextType] = None,
                 level: Optional[TokenLevel] = None, state: Optional[TokenState] = None,
                 target: Optional[ObjectInformation] = None,
                 attribute_form: Optional[ObjectReference] = None):
        self.value = value
        self.type = type
        self.level = level
        self.state = state
        self.target = target
        self.attrtibute_form = attribute_form

    _FROM_DICT_MAP = {
        "type": TokenTextType,
        "level": TokenLevel,
        "state": TokenState,
        "target": ObjectInformation.from_dict,
        "attribute_form": ObjectReference.from_dict
    }



import time
from typing import List, Optional, Union

from packaging import version
import pandas as pd
from tqdm.auto import tqdm

from mstrio import config
from mstrio.api import datasets
from mstrio.object_management.search_operations import full_search, SearchPattern
from mstrio.connection import Connection
from mstrio.utils import helper
from mstrio.utils.encoder import Encoder
from mstrio.utils.entity import CertifyMixin, ObjectSubTypes
from mstrio.utils.model import Model

from .cube import _Cube


def list_super_cubes(connection: Connection, name_begins: Optional[str] = None,
                     to_dictionary: bool = False, limit: Optional[int] = None,
                     **filters) -> Union[List["SuperCube"], List[dict]]:
    """Get list of SuperCube objects or dicts with them.
    Optionally filter cubes by specifying 'name_begins'.

    Optionally use `to_dictionary` to choose output format.

    Wildcards available for 'name_begins':
        ? - any character
        * - 0 or more of any characters
        e.g name_begins = ?onny wil return Sonny and Tonny

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        name_begins (string, optional): characters that the cube name must begin
            with
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns SuperCube objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        **filters: Available filter parameters: ['id', 'name', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'owner',
            'ext_type', 'view_media', 'certified_info']

    Returns:
        list with SuperCubes or list of dictionaries
    """
    connection._validate_project_selected()
    objects_ = full_search(connection, object_types=ObjectSubTypes.SUPER_CUBE,
                           project=connection.project_id, name=name_begins,
                           pattern=SearchPattern.BEGIN_WITH, limit=limit, **filters)
    if to_dictionary:
        return objects_
    else:
        return [SuperCube.from_dict(obj_, connection) for obj_ in objects_]


class SuperCube(_Cube, CertifyMixin):
    """Manage multiple table cube (MTDI aka Super Cube) - according to
    EnumDSSXMLObjectSubTypes its subtype is 779 (DssXmlSubTypeReportEmmaCube).
    It inherits all properties from Cube.

    Attributes:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        id: Identifier of a pre-existing cube.
        instance_id (str): Identifier of a cube instance if already initialized,
            None by default.
        name: Name of the SuperCube.
        description: Description given to the SuperCube.
        id: Unique identifier for the super cube. Used to update a
            existing super cube or generated after creating a new super cube.
            (deprecated)
        upload_body (string): upload body of super cube
        session_id (string): ID of session used for uploading and publishing
            super cube
        size(integer): size of cube
        status(integer): status of cube
        path(string): full path of the cube on environment
        owner_id(string): ID of cube's owner
        attributes(list): all attributes of cube
        metrics(list): all metrics of cube
        attr_elements(list): all attributes elements of cube
        selected_attributes(list): IDs of filtered attributes
        selected_metrics(list): IDs of filtered metrics
        selected_attr_elements(list): IDs of filtered attribute elements
        dataframe(object): content of a cube extracted into a
            Pandas `DataFrame`
        table_definition
    """
    _OBJECT_SUBTYPE = ObjectSubTypes.SUPER_CUBE.value
    __VALID_POLICY = ['add', 'update', 'replace', 'upsert']
    __MAX_DESC_LEN = 250

    def __init__(self, connection: Connection, id: Optional[str] = None,
                 name: Optional[str] = None, description: Optional[str] = None,
                 instance_id: Optional[str] = None, progress_bar: bool = True,
                 parallel: bool = True):
        """Initialize super cube.

        When creating new super cube, provide its `name` and an optional
        `description`.
        When updating a pre-existing super cube, provide its `id`. This
        identifier will be then used to initialize super cube.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str, optional): Identifier of a pre-existing super cube.
                Used when updating a pre-existing super cube.
            name (str): Name of the super cube.
            description (str, optional): Description of the super cube. Must be
                less than or equal to 250 characters.
            instance_id (str): Identifier of an instance if cube instance has
                been already initialized, NULL by default.
            progress_bar(bool, optional): If True (default), show the upload
                progress bar.
            parallel (bool, optional): If True (default), utilize optimal number
                of threads to increase the download speed. If False, this
                feature will be disabled.
        """
        if name is not None:
            self.__check_param_str(name, msg="Super cube name should be a string.")
            self.__check_param_len(
                name,
                msg="Super cube name should be <= {} characters.".format(self.__MAX_DESC_LEN),
                max_length=self.__MAX_DESC_LEN)

        if description is not None:
            self.__check_param_str(description, msg="Super cube description should be a string.")
            self.__check_param_len(
                description, msg="Super cube description should be <= {} characters.".format(
                    self.__MAX_DESC_LEN), max_length=self.__MAX_DESC_LEN)

        connection._validate_project_selected()

        if id is not None:
            super().__init__(connection=connection, id=id, instance_id=instance_id,
                             parallel=parallel, progress_bar=progress_bar)
        else:
            self._init_variables(connection=connection, name=name, description=description,
                                 progress_bar=progress_bar, parallel=parallel)

    def _init_variables(self, **kwargs):
        super()._init_variables(**kwargs)
        self._tables = []
        self._session_id = None
        self._folder_id = None
        self.__upload_body = None

    def add_table(self, name, data_frame, update_policy, to_metric=None, to_attribute=None):
        """Add a `Pandas.DataFrame` to a collection of tables which are later
        used to populate the MicroStrategy super cube with data.

        Args:
            name (str): Logical name of the table that is visible to users of
                the super cube in MicroStrategy.
            data_frame (:obj:`pandas.DataFrame`): Pandas DataFrame to add or
                update.
            update_policy (str): Update operation to perform. One of 'add'
                (inserts new, unique rows), 'update' (updates data in existing
                rows and columns), 'upsert' (updates existing data and inserts
                new rows), or 'replace' (replaces the existing data with new
                data).
            to_metric (optional, :obj:`list` of str): By default, Python numeric
                data types are treated as metrics while character and date types
                are treated as attributes. For example, a column of integer-like
                strings ("1", "2", "3") would, by default, be an attribute. If
                the intent is to format this data as a metric, provide the
                respective column name as a string in a list to the `to_metric`
                parameter.
            to_attribute (optional, :obj:`list` of str): Logical opposite of
                `to_metric`. Helpful for formatting an integer-based row
                identifier as a primary key in the super cube.
        """
        update_policy = update_policy.lower()
        version_ok = version.parse(self._connection.iserver_version) >= version.parse("11.2.0300")

        if not isinstance(data_frame, pd.DataFrame):
            raise TypeError("`data_frame` parameter must be a valid `Pandas.DataFrame`.")
        if not helper.check_duplicated_column_names(data_frame):
            raise ValueError(
                "`DataFrame` column names need to be unique for each table in the super cube.")
        if update_policy not in self.__VALID_POLICY:
            raise ValueError(f"Update policy must be one of {self.__VALID_POLICY}.")
        if self._id is None and update_policy != "replace" and version_ok:
            raise ValueError(
                "Update policy has to be 'replace' if a super cube is created or overwritten.")
        if to_attribute and to_metric and any(col in to_attribute for col in to_metric):
            raise ValueError(
                "Column name(s) present in `to_attribute` also present in `to_metric`.")

        table = {"table_name": name, "data_frame": data_frame, "update_policy": update_policy}

        if to_attribute is not None:
            if any((col for col in to_attribute if col not in data_frame.columns)):
                raise ValueError(
                    "Column name(s) in `to_attribute` were not found in `DataFrame.columns`.")
            else:
                table["to_attribute"] = to_attribute

        if to_metric is not None:
            if any((col for col in to_metric if col not in data_frame.columns)):
                raise ValueError(
                    "Column name(s) in `to_metric` were not found in `DataFrame.columns`.")
            else:
                table["to_metric"] = to_metric

        self._tables.append(table)

    def create(self, folder_id: Optional[str] = None, auto_upload: bool = True,
               auto_publish: bool = True, chunksize: int = 100000) -> None:
        """Create a new super cube and initialize cube object after successful
        creation. This function does not return new super cube, but it updates
        object inplace.

        Args:
            folder_id (str, optional): ID of the shared folder in which the
                super cube will be created. If `None`, defaults to the user's
                My Reports folder.
            auto_upload (bool, optional): If True, automatically uploads the
                data to the I-Server. If False, simply creates the super cube
                definition but does not upload data to it.
            auto_publish (bool, optional): If True, automatically publishes the
                data used to create the super cube definition. If False, simply
                creates the super cube but does not publish it. To publish the
                super cube, data has to be uploaded first.
            chunksize (int, optional): Number of rows to transmit to the
                I-Server with each request when uploading.
        """
        if auto_publish and not auto_upload:
            raise ValueError(
                "Data needs to be uploaded to the I-Server before the super cube can be published."
            )

        if folder_id is not None:
            self._folder_id = folder_id
        else:
            self._folder_id = ""

        # generate model of the super cube
        self.__build_model()

        # makes request to create the super cube
        response = datasets.create_multitable_dataset(connection=self._connection,
                                                      body=self.__model)
        self._set_object(**response.json())

        if config.verbose:
            print("Created super cube '{}' with ID: '{}'.".format(*[self.name, self._id]))

        if auto_upload:
            self.update(chunksize=chunksize, auto_publish=auto_publish)

        # after creating super cube fetch definition and create filter object
        self._get_definition()

    def update(self, chunksize: int = 100000, auto_publish: bool = True):
        """Updates a super cube with new data.

        Args:
            chunksize (int, optional): Number of rows to transmit to the server
                with each request.
            auto_publish: If True, automatically publishes the data used to
                update the super cube definition to the super cube. If False,
                simply updates the super cube but does not publish it.
        """

        # form request body and create a session for data uploads
        self.__form_upload_body()
        response = datasets.upload_session(connection=self._connection, id=self._id,
                                           body=self.__upload_body)

        response_json = response.json()
        self._session_id = response_json['uploadSessionId']

        # upload each table
        for ix, _table in enumerate(self._tables):

            _df, _name = _table["data_frame"], _table["table_name"]

            # break the data up into chunks using a generator
            chunks = (_df[i:i + chunksize] for i in range(0, _df.shape[0], chunksize))

            total = _df.shape[0]

            # Count the number of iterations
            it_total = int(total / chunksize) + (total % chunksize != 0)

            pbar = tqdm(chunks, total=it_total, disable=(not self._progress_bar))
            for index, chunk in enumerate(pbar):
                pbar.set_description("Uploading {}/{}".format(ix + 1, len(self._tables)))

                # base64 encode the data
                encoder = Encoder(data_frame=chunk, dataset_type='multi')
                b64_enc = encoder.encode

                # form body of the request
                body = {
                    "tableName": _name,
                    "index": index + 1,
                    "data": b64_enc,
                }

                # make request to upload the data
                response = datasets.upload(
                    connection=self._connection,
                    id=self._id,
                    session_id=self._session_id,
                    body=body,
                )

                if not response.ok:
                    # on error, cancel the previously uploaded data
                    datasets.publish_cancel(connection=self._connection, id=self._id,
                                            session_id=self._session_id)

                pbar.set_postfix(rows=min((index + 1) * chunksize, total))
            pbar.close()
        self._tables = []

        # if desired, automatically publish the data to the new super cube
        if auto_publish:
            self.publish()

    def save_as(self, name: str, description: Optional[str] = None,
                folder_id: Optional[str] = None, table_name: Optional[str] = None) -> "SuperCube":
        """Creates a new single-table cube with the data frame stored in the
        SuperCube instance `SuperCube.dataframe`.

        Args:
            name(str): Name of cube.
            description(str): Description of the cube.
            folder_id (str, optional): ID of the shared folder that the super
                cube should be created within. If `None`, defaults to the user's
                My Reports folder.
            table_name (str, optional): Name of the table. If None (default),
                the first table name of the original cube will be used.
        """
        if len(self._table_names) > 1:
            helper.exception_handler(msg="""This feature works only for the single-table cubes.
                                            \rTo export multi-table cube use `create` method.""")
        elif self._dataframe is None:
            raise ValueError(
                "Current cube has to be fetched with `to_dataframe` method first to create a copy."
            )
        else:
            if table_name is None:
                table_name = self._table_names[0]["name"]
            new_cube = SuperCube(self._connection, name=name, description=description)
            new_cube.add_table(name=table_name, data_frame=self.dataframe, update_policy="replace")
            new_cube.create(folder_id=folder_id)
            return new_cube

    def publish(self) -> bool:
        """Publish the uploaded data to the selected super cube.

        Note:
            A super cube can be published just once.

        Returns:
            True if the data was published successfully, else False.
        """

        response = datasets.publish(connection=self._connection, id=self._id,
                                    session_id=self._session_id)

        if not response.ok:
            # on error, cancel the previously uploaded data
            datasets.publish_cancel(connection=self._connection, id=self._id,
                                    session_id=self._session_id)
            return False

        status = 6  # default initial status
        while status != 1:
            status = self.publish_status()['status']
            time.sleep(1)
            if status == 1:
                # clear instance_id to force new instance creation
                self.instance_id = None
                if config.verbose:
                    print(f"Super cube '{self.name}' published successfully.")
                return True

    def publish_status(self):
        """Check the status of data that was uploaded to a super cube.

        Returns:
            status: The status of the publication process as a dictionary. In
                the 'status' key, "1" denotes completion.
        """
        if not self._session_id:
            raise AttributeError("No upload session created.")
        else:
            response = datasets.publish_status(connection=self._connection, id=self._id,
                                               session_id=self._session_id)
            return response.json()

    def upload_status(self, connection: Connection, id: str, session_id: str):
        """Check the status of data that was uploaded to a super cube.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Identifier of a pre-existing super cube.
            session_id (str): Identifier of the server session used for
                collecting uploaded data.
        """
        # TODO not sure why we have this functionality twice
        response = datasets.publish_status(connection=connection, id=id, session_id=session_id)

        helper.response_handler(response=response,
                                msg="Publication status for super cube with ID: '{}':".format(id),
                                throw_error=False)

    def __build_model(self):
        """Create json representation of the super cube."""

        # generate model
        model = Model(tables=self._tables, name=self.name, description=self.description,
                      folder_id=self._folder_id)
        self.__model = model.get_model()

    def __form_upload_body(self):
        """Form request body for creating an upload session for data
        uploads."""

        # generate body string
        body = {
            "tables": [{
                "name": tbl["table_name"],
                "updatePolicy": tbl["update_policy"],
                "columnHeaders": list(tbl["data_frame"].columns)
            } for tbl in self._tables]
        }
        self.__upload_body = body

    @staticmethod
    def __check_param_len(param, msg, max_length):
        if len(param) > max_length:
            raise ValueError(msg)
        else:
            return True

    @staticmethod
    def __check_param_str(param, msg):
        if not isinstance(param, str):
            raise TypeError(msg)
        else:
            return True

    @property
    def upload_body(self):
        return self.__upload_body

    @property
    def session_id(self) -> str:
        return self._session_id



"""This is the demo script to show how administrator can monitor and manage jobs

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.

Note: Some functionality is missing when working with 11.3.2 I-Server
    versions. To use all functionality 11.3.3 and up will be required.
    There are also some differences between 11.3.2 and 11.3.3+ Job
    attributes values.
"""

from mstrio import connection
from mstrio.connection import Connection
from mstrio.server.project import Project
from mstrio.server.job_monitor import (Job, JobStatus, JobType, kill_all_jobs, kill_jobs,
                                       list_jobs, ObjectType, PUName)
from mstrio.users_and_groups.user import User

base_url = "https://<>/MicroStrategyLibrary/api"
username = "some_username"
password = "some_password"
conn = Connection(base_url, username, password, login_mode=1)

# get project by name
project = Project(connection=conn, name="MicroStrategy Tutorial")
# get user by name
john = User(connection=conn, name="John Smith")

# instanstiate an existing job using constructor
job = Job(conn, id="55AA293811EAE2F2EC7D0080EF25A593")  # example job ID
# get properties for the first job on the list
job.list_properties()
# kill the job
job.kill()

# get list of `Job` objects for all jobs on the environment
jobs = list_jobs(conn)

# get list of dicts representing job information for all jobs on the environment
jobs = list_jobs(conn, to_dictionary=True)

# get a list of `Job` objects filtered by job status, project and object type
list_jobs(conn, status=JobStatus.LOADING_PROMPT, project=project, object_type=ObjectType.CUBE)

# get a list of `Job` objects filtered by job type and job owner
list_jobs(conn, type=JobType.INTERACTIVE, user=john)

# get a list of `Job` objects filtered by elapsed time and memory usage
# NOTE: memory_usage filter works for 11.3.3+ I-Server versions
slow_jobs = list_jobs(conn, elapsed_time="gt:10000", memory_usage="gt:500")

# kill jobs by passing either Job objects or ids
result = kill_jobs(conn, slow_jobs)
# kill_jobs return Success, PartialSuccess or MSTRException, you can check which
# jobs were killed and which were not, and why in case of PartialScucess
result.succeeded
result.failed
# you can easily evaluate if all passed jobs were killed using bool() evaluation
# on result, which will return True if result is Success and False if result
# is PartialSuccess
succesfully_killed_passed_jobs = bool(result)
if result:
    print("Successfuly killed jobs.")
    # perform other automation actions

# kill jobs running over 5 hours
time_hours = 5
elapsed_t = time_hours * 60**2  # convert to seconds (for 11.3.2 I-Server version)
if connection.iserver_version >= '11.3.0300':
    elapsed_t = 1000 * elapsed_t  # convert to milliseconds (for 11.3.3+ I-Server versions)
elapsed_t = f'gt:{elapsed_t}'  # correct form filter (valid operators: 'gt:' and 'lt:')
kill_all_jobs(conn, elapsed_time=elapsed_t, force=True)

# kill filtered jobs similarly to list_jobs() using single function
result = kill_all_jobs(conn, pu_name=PUName.SQL_ENGINE, memory_usage="gt:800")
# like kill_jobs, kill_all_jobs return Success, PartialSuccess or MSTRException
result.succeeded



# flake8: noqa
from mstrio.project_objects.datasets import *
from mstrio.project_objects.document import Document, list_documents, list_documents_across_projects
from mstrio.project_objects.dossier import Dossier, list_dossiers, list_dossiers_across_projects
from mstrio.project_objects.library import Library
from mstrio.project_objects.report import list_reports, Report
from mstrio.utils.helper import deprecation_warning

deprecation_warning(
    'mstrio.application_objects',
    'mstrio.project_objects',
    '11.3.4.101',  # NOSONAR
    False)



__title__ = "mstrio-py"
__version__ = "11.3.3.102"  # NOSONAR
__dev_version__ = 1027
__license__ = "Apache License 2.0"
__description__ = "Python interface for the MicroStrategy REST API"
__author__ = "MicroStrategy"
__author_email__ = "pkowal@microstrategy.com"



from typing import List, Optional
from enum import Enum

from mstrio.utils.helper import Dictable
from mstrio.utils import helper


class ScheduleEnums:
    """Object representations of recurrence information of a time-based schedule
    """

    class DaysOfWeek(Enum):
        MONDAY = 'monday'
        TUESDAY = 'tuesday'
        WEDNESDAY = 'wednesday'
        THURSDAY = 'thursday'
        FRIDAY = 'friday'
        SATURDAY = 'saturday'
        SUNDAY = 'sunday'
        NONE = None

    class WeekOffset(Enum):
        FIRST = 'first'
        SECOND = 'second'
        THIRD = 'third'
        FOURTH = 'fourth'
        LAST = 'last'
        NONE = None

    class RecurrencePattern(Enum):
        DAILY = 'daily'
        WEEKLY = 'weekly'
        MONTHLY = 'monthly'
        YEARLY = 'yearly'

    class ExecutionPattern(Enum):
        ONCE = 'once'
        REPEAT = 'repeat'

    class DailyPattern(Enum):
        DAY = 'day'
        WEEKDAY = 'weekday'
        NONE = 'none'

    class MonthlyPattern(Enum):
        DAY = 'day'
        DAY_OF_WEEK = 'day_of_week'
        WEEKDAY = 'weekday'
        LAST_DAY = 'last_day'
        DAYS_OF_MONTH = 'days_of_month'
        NONE = None

    class WeekdayOffset(Enum):
        FIRST = 'first'
        LAST = 'last'
        NONE = None

    class YearlyPattern(Enum):
        DAY = 'day'
        DAYOFWEEK = 'day_of_week'
        NONE = None


class ScheduleTime(Dictable):
    """Object representation of details of time-based schedule"""

    class Execution(Dictable):
        """Object representation of the execution information of the schedule"""

        def __init__(self, execution_pattern: ScheduleEnums.ExecutionPattern = ScheduleEnums
                     .ExecutionPattern.ONCE, execution_time: str = None, start_time: str = None,
                     stop_time: str = None, repeat_interval: int = 1):
            """Set attributes representing execution information of the schedule

            Args:
                execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                    The execution pattern of the schedule.
                    Possible values are ONCE or REPEAT.
                    Defaults to ScheduleEnums.ExecutionPattern.ONCE.
                execution_time (str, optional):
                    The execution time of the execution day, if
                    execution_pattern is once. Format should be HH:mm:ss.
                    Defaults to ''.
                start_time (str, optional):
                    The start time of the execution day, if execution_pattern
                    is repeat. Format should be HH:mm:ss. Defaults to ''.
                stop_time (str, optional):
                    The stop time of the execution day, if execution_pattern is
                    repeat. Format should be HH:mm:ss. Defaults to ''.
                repeat_interval (int, optional):
                    The repeat interval of minutes of the execution day, if
                    executionPattern is repeat. Defaults to 1.
            Raises:
                ValueError: when execution_repeat_interval = None if
                execution_pattern == REPEAT
            """
            if (execution_pattern == ScheduleEnums.ExecutionPattern.REPEAT
                    and repeat_interval is None):
                helper.exception_handler(
                    msg=('Value error: execution_repeat_interval '
                         'cannot be None, '
                         'for execution_pattern == ExecutionPattern.Repeat'),
                    exception_type=ValueError)
            if execution_pattern == ScheduleEnums.ExecutionPattern.REPEAT:
                self.start_time = start_time
                self.stop_time = stop_time
                self.repeat_interval = repeat_interval
            else:
                self.execution_time = execution_time
            self.execution_pattern = execution_pattern

        _FROM_DICT_MAP = {
            'execution_pattern': ScheduleEnums.ExecutionPattern,
        }

    class Daily(Dictable):
        """Object representation of daily recurrence information of the schedule
        """

        def __init__(self,
                     daily_pattern: ScheduleEnums.DailyPattern = ScheduleEnums.DailyPattern.DAY,
                     repeat_interval: int = 1):
            """Set attributes representing daily recurrence information of the
            schedule

            Args:
                daily_pattern (ScheduleEnums.DailyPattern, optional):
                    The daily recurrence pattern of the schedule.Possible values
                    are DAY, WEEKDAY.Defaults to ScheduleEnums.DailyPattern.DAY.
                repeat_interval (int, optional):
                    The repeat interval of days of daily schedule,
                    if daily_pattern is DAY. Defaults to 1.
            """
            self.daily_pattern = daily_pattern
            if daily_pattern == ScheduleEnums.DailyPattern.DAY:
                self.repeat_interval = repeat_interval

        _FROM_DICT_MAP = {'daily_pattern': ScheduleEnums.DailyPattern}

    class Weekly(Dictable):
        """Object representation of weekly recurrence information of the
        schedule"""

        def __init__(self, days_of_week: List[ScheduleEnums.DaysOfWeek] = None,
                     repeat_interval: int = 1):
            """Set attributes representing weekly recurrence information of the
            schedule.

            Args:
                days_of_week (List[ScheduleEnums.DaysOfWeek], optional):
                    The days of week of weekly schedule. Possible values are:
                    MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY,
                    SUNDAY. Defaults to None.
                repeat_interval (int, optional):
                    The repeat interval of weeks of weekly schedule.
                    Defaults to 1.
            """
            self.days_of_week = days_of_week or []
            self.repeat_interval = repeat_interval

        _FROM_DICT_MAP = {'days_of_week': [ScheduleEnums.DaysOfWeek]}

    class Monthly(Dictable):
        """Object representation of monthly recurrence information of the
        schedule"""

        def __init__(self, monthly_pattern: ScheduleEnums.MonthlyPattern, repeat_interval: int,
                     day: Optional[int] = None,
                     week_offset: Optional[ScheduleEnums.WeekOffset] = None,
                     day_of_week: Optional[ScheduleEnums.DaysOfWeek] = None,
                     weekday_offset: Optional[ScheduleEnums.WeekdayOffset] = None,
                     days_of_month: Optional[str] = None):
            """Set attributes representing monthly recurrence information of the
            schedule.

            Args:
                monthly_pattern (ScheduleEnums.MonthlyPattern):
                    The monthly recurrence pattern of the schedule. Possible
                    values are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY,
                    DAYS_OF_MONTH.
                repeat_interval (int):
                    The repeat interval of months of monthly schedule.
                day (Optional[int], optional):
                    The day in month of monthly schedule, if monthly_pattern is
                    DAY. Defaults to None.
                week_offset (Optional[ScheduleEnums.WeekOffset], optional):
                    The week offset in month of monthly schedule, if
                    monthly_pattern is DAY_OF_WEEK. Possible values are:
                    FIRST, SECOND, THIRD, FOURTH, LAST. Defaults to None.
                day_of_week (Optional[ScheduleEnums.DaysOfWeek], optional):
                    The day of week in month of monthly schedule, if
                    monthly_pattern is DAY_OF_WEEK. Possible values are:
                    MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY,
                    SUNDAY. Defaults to None.
                weekday_offset (Optional[ScheduleEnums.WeekdayOffset], optional)
                    The weekday offset in month of monthly schedule,
                    if monthly_pattern is WEEKDAY. Possible values are: FIRST,
                    LAST. Defaults to None.
                days_of_month (Optional[str], optional):
                    The days of month of monthly schedule, if monthly_pattern
                    is DAYS_OF_MONTH. Must be provided as a list of one or more
                    stringified digits (from '1' to '31'). Defaults to None.
            """

            self.monthly_pattern = monthly_pattern
            self.repeat_interval = repeat_interval

            if monthly_pattern == ScheduleEnums.MonthlyPattern.DAY:
                self.day = day
            elif monthly_pattern == ScheduleEnums.MonthlyPattern.DAY_OF_WEEK:
                self.week_offset = week_offset
                self.day_of_week = day_of_week
            elif monthly_pattern == ScheduleEnums.MonthlyPattern.WEEKDAY:
                self.weekday_offset = weekday_offset
            elif monthly_pattern == ScheduleEnums.MonthlyPattern.DAYS_OF_MONTH:
                self.days_of_month = days_of_month

        _FROM_DICT_MAP = {
            'monthly_pattern': ScheduleEnums.MonthlyPattern,
            'week_offset': ScheduleEnums.WeekOffset,
            'day_of_week': ScheduleEnums.DaysOfWeek,
            'weekday_offset': ScheduleEnums.WeekdayOffset
        }

    class Yearly(Dictable):
        """Object representation of yearly recurrence information of the
        schedule"""

        def __init__(self,
                     yearly_pattern: ScheduleEnums.YearlyPattern = ScheduleEnums.YearlyPattern.DAY,
                     month: int = 1, day: int = 1, week_offset: ScheduleEnums.WeekOffset = None,
                     day_of_week: ScheduleEnums.DaysOfWeek = None):
            """Set attributes representing yearly recurrence information of the
            schedule

            Args:
                yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                    The yearly recurrence pattern of the schedule. Possible
                    values are DAY, DAY_OF_WEEK.
                    Defaults to ScheduleEnums.YearlyPattern.DAY.
                month (int, optional):
                    The month in year of yearly schedule. Defaults to 1.
                day (int, optional):
                    The day in month of yearly schedule, if yearly_pattern
                    is DAY. Defaults to 1.
                week_offset (ScheduleEnums.WeekOffset, optional):
                    The week offset in year of yearly schedule, if
                    yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                    SECOND, THIRD, FOURTH, LAST. Defaults to None.
                day_of_week (ScheduleEnums.DaysOfWeek, optional):
                    The day of week in year of yearly schedule,
                    if yearly_pattern is DAY_OF_WEEK. Possible values are:
                    MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY,
                    SUNDAY. Defaults to None.
            """
            self.yearly_pattern = yearly_pattern
            self.month = month
            if yearly_pattern == ScheduleEnums.YearlyPattern.DAY:
                self.day = day
            if yearly_pattern == ScheduleEnums.YearlyPattern.DAYOFWEEK:
                self.week_offset = week_offset
                self.day_of_week = day_of_week

        _FROM_DICT_MAP = {
            'yearly_pattern': ScheduleEnums.YearlyPattern,
            'week_offset': ScheduleEnums.WeekOffset,
            'day_of_week': ScheduleEnums.DaysOfWeek
        }

    _FROM_DICT_MAP = {
        'recurrence_pattern': ScheduleEnums.RecurrencePattern,
        'days_of_week': ScheduleEnums.DaysOfWeek,
        'execution': Execution.from_dict,
        'daily': Daily.from_dict,
        'weekly': Weekly.from_dict,
        'monthly': Monthly.from_dict,
        'yearly': Yearly.from_dict
    }

    def __init__(self, recurrence_pattern: ScheduleEnums.RecurrencePattern = None,
                 execution: Execution = None, daily: Daily = None, weekly: Weekly = None,
                 monthly: Monthly = None, yearly: Yearly = None):
        """Set attributes representing details of the time-based schedule.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule. Possible values are
                DAILY, WEEKLY, MONTHLY, YEARLY. Defaults to None.
            execution (Execution, optional):
                Object representing the execution information of the schedule.
                Defaults to None.
            daily (Daily, optional):
                Object representing daily recurrence information of the
                schedule. Defaults to None.
            weekly (Weekly, optional):
                Object representing weekly recurrence information of the
                schedule. Defaults to None.
            monthly (Monthly, optional):
                Object representing monthly recurrence information of the
                schedule. Defaults to None.
            yearly (Yearly, optional):
                Object representing yearly recurrence information of the
                schedule. Defaults to None.
        """

        self.recurrence_pattern = recurrence_pattern
        self.execution = execution
        if daily:
            self.daily = daily
        elif weekly:
            self.weekly = weekly
        elif monthly:
            self.monthly = monthly
        elif yearly:
            self.yearly = yearly

    @classmethod
    def from_details(
            cls, recurrence_pattern: ScheduleEnums.RecurrencePattern,
            execution_pattern: ScheduleEnums.ExecutionPattern,
            execution_time: Optional[str] = None, start_time: Optional[str] = None,
            stop_time: Optional[str] = None, execution_repeat_interval: Optional[int] = None,
            daily_pattern: Optional[ScheduleEnums.DailyPattern] = ScheduleEnums.DailyPattern.NONE,
            repeat_interval: Optional[int] = None,
            days_of_week: Optional[List[ScheduleEnums.DaysOfWeek]] = None,
            day: Optional[int] = None, month: Optional[int] = None,
            week_offset: Optional[ScheduleEnums.WeekOffset] = ScheduleEnums.WeekOffset.NONE,
            day_of_week: Optional[ScheduleEnums.DaysOfWeek] = ScheduleEnums.DaysOfWeek.NONE,
            weekday_offset: Optional[str] = None, days_of_month: Optional[List[str]] = None,
            monthly_pattern: Optional[
                ScheduleEnums.MonthlyPattern] = ScheduleEnums.MonthlyPattern.NONE,
            yearly_pattern: Optional[
                ScheduleEnums.YearlyPattern] = ScheduleEnums.YearlyPattern.NONE):
        """
        Uses provided properties to create object representation of details of a
        time-based schedule.

        This object is used as 'time' part of request body and is needed for
        creating and updating schedule.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule.
                Possible values are DAILY, WEEKLY, MONTHLY, YEARLY
                Defaults to None.
            execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                The execution pattern of the schedule.
                Possible values are ONCE, REPEAT. Defaults to None.
            execution_time (str, optional):
                The execution time of the execution day, if execution_pattern
                is ONCE. Format should be HH:mm:ss. Defaults to ''.
            start_time (str, optional):
                The start time of the execution day, if execution_pattern
                is REPEAT. Format should be HH:mm:ss. Defaults to ''.
            stop_time (str, optional):
                The stop time of the execution day, if execution_pattern
                is REPEAT. Format should be HH:mm:ss.
                Defaults to ''.
            execution_repeat_interval (int, optional):
                The repeat interval of minutes of the execution day, if
                execution_pattern is REPEAT. Defaults to None.
            daily_pattern (ScheduleEnums.DailyPattern, optional):
                The daily recurrence pattern of the schedule.
                Possible values are DAY, WEEKDAY.
                Defaults to None.
            repeat_interval (int, optional):
                The repeat interval of days of daily schedule, if daily_pattern
                is DAY or, The repeat interval of weeks of weekly schedule or,
                The repeat interval of months of monthly schedule.
                Defaults to None.
            day (int, optional):
                The day in month of monthly schedule, if monthly_pattern is DAY
                or, The day in month of yearly schedule, if yearly_pattern is
                DAY. Defaults to None.
            month (int, optional):
                The month in year of yearly schedule.
                Defaults to None.
            week_offset (ScheduleEnums.WeekOffset, optional):
                The week offset in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The week offset in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                SECOND, THIRD, FOURTH, LAST. Defaults to None.
            day_of_week (ScheduleEnums.DaysOfWeek, optional):
                The days of week of weekly schedule or, The day of week in month
                of monthly schedule, if monthly_pattern is DAY_OF_WEEK or,
                The day of week in year of yearly schedule, if yearly_pattern is
                DAY_OF_WEEK. Possible values are: MONDAY, TUESDAY, WEDNESDAY,
                THURSDAY, FRIDAY, SATURDAY, SUNDAY. Defaults to None.
            weekday_offset (ScheduleEnums.WeekdayOffset, optional):
                The weekday offset in month of monthly schedule, if
                monthly_pattern is WEEKDAY. Possible values are: FIRST, LAST.
                Defaults to None.
            days_of_month (List[str], optional):
                The days of month of monthly schedule, if monthly_pattern is
                DAYS_OF_MONTH. Must be provided as a list of one or more
                stringified digits (from '1' to '31'). Defaults to None.
            monthly_pattern (ScheduleEnums.MonthlyPattern, optional):
                The monthly recurrence pattern of the schedule. Possible values
                are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY, DAYS_OF_MONTH.
                Defaults to None.
            yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                The yearly recurrence pattern of the schedule. Possible values
                are DAY, DAY_OF_WEEK. Defaults to None.
        Returns:
            ScheduleTime: Object representation of details of time-based
            schedule.
        """

        if execution_pattern == ScheduleEnums.ExecutionPattern.ONCE:
            execution = cls.Execution.from_dict({
                'execution_pattern': execution_pattern,
                'execution_time': execution_time
            })
        elif execution_pattern == ScheduleEnums.ExecutionPattern.REPEAT:
            execution = cls.Execution.from_dict({
                'execution_pattern': execution_pattern,
                'start_time': start_time,
                'stop_time': stop_time,
                'repeat_interval': execution_repeat_interval
            })
        else:
            helper.exception_handler(msg='Error: Wrong value of execution_pattern',
                                     exception_type=ValueError)

        if recurrence_pattern == ScheduleEnums.RecurrencePattern.DAILY:
            daily = cls.Daily.from_dict({
                'daily_pattern': daily_pattern,
                'repeat_interval': repeat_interval
            })
            return cls(recurrence_pattern=recurrence_pattern, execution=execution, daily=daily)
        elif recurrence_pattern == ScheduleEnums.RecurrencePattern.WEEKLY:
            if days_of_week and type(days_of_week) is not list:
                helper.exception_handler(msg='Error: days_of_week must be provided as a list',
                                         exception_type=ValueError)
            weekly = cls.Weekly.from_dict({
                'repeat_interval': repeat_interval,
                'days_of_week': days_of_week
            })
            return cls(recurrence_pattern=recurrence_pattern, execution=execution, weekly=weekly)
        elif recurrence_pattern == ScheduleEnums.RecurrencePattern.MONTHLY:
            monthly = cls.Monthly.from_dict({
                "monthly_pattern": monthly_pattern,
                "repeatInterval": repeat_interval,
                "day": day,
                "week_offset": week_offset,
                "day_of_week": day_of_week,
                "WeekdayOffset": weekday_offset,
                "days_of_month": days_of_month
            })
            return cls(recurrence_pattern=recurrence_pattern, execution=execution, monthly=monthly)
        elif recurrence_pattern == ScheduleEnums.RecurrencePattern.YEARLY:
            yearly = cls.Yearly.from_dict({
                'yearly_pattern': yearly_pattern,
                'month': month,
                'day': day,
                'week_offset': week_offset,
                'day_of_week': day_of_week
            })
            return cls(recurrence_pattern=recurrence_pattern, execution=execution, yearly=yearly)
        else:
            helper.exception_handler(msg='Error: Wrong value of recurrence_pattern',
                                     exception_type=ValueError)

    def update_properties(
            self, recurrence_pattern: ScheduleEnums.RecurrencePattern = None,
            execution_pattern: ScheduleEnums.ExecutionPattern = None, execution_time: str = None,
            start_time: str = None, stop_time: str = None, execution_repeat_interval: int = None,
            daily_pattern: ScheduleEnums.DailyPattern = None, repeat_interval: int = None,
            days_of_week: List[ScheduleEnums.DaysOfWeek] = None, day: int = None,
            month: int = None, week_offset: ScheduleEnums.WeekOffset = None,
            day_of_week: ScheduleEnums.DaysOfWeek = None,
            weekday_offset: ScheduleEnums.WeekdayOffset = None, days_of_month: List[str] = None,
            monthly_pattern: ScheduleEnums.MonthlyPattern = None,
            yearly_pattern: ScheduleEnums.YearlyPattern = None):
        """
        Updates ScheduleTime object according to provided parameters. If a
        provided parameter is not None then update it, else use its current
        value.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule. Possible values are
                DAILY, WEEKLY, MONTHLY, YEARLY. Defaults to None.
            execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                The execution pattern of the schedule. Possible values are ONCE,
                REPEAT. Defaults to None.
            execution_time (str, optional):
                The execution time of the execution day, if execution_pattern is
                ONCE. Format should be HH:mm:ss. Defaults to ''.
            start_time (str, optional):
                The start time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to ''.
            stop_time (str, optional):
                The stop time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to ''.
            execution_repeat_interval (int, optional):
                The repeat interval of minutes of the execution day, if
                execution_pattern is REPEAT. Defaults to None.
            daily_pattern (ScheduleEnums.DailyPattern, optional):
                The daily recurrence pattern of the schedule. Possible values
                are DAY, WEEKDAY. Defaults to None.
            repeat_interval (int, optional):
                The repeat interval of days of daily schedule, if daily_pattern
                is DAY. Defaults to None.
            day (int, optional):
                The day in month of monthly schedule, if monthly_pattern is DAY
                or The day in month of yearly schedule, if yearly_pattern is
                DAY. Defaults to None.
            month (int, optional):
                The month in year of yearly schedule. Defaults to None.
            week_offset (ScheduleEnums.WeekOffset, optional):
                The week offset in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The week offset in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                SECOND, THIRD, FOURTH, LAST. Defaults to None.
            day_of_week (ScheduleEnums.DaysOfWeek, optional):
                The days of week of weekly schedule or, The day of week in month
                of monthly schedule, if monthly_pattern is DAY_OF_WEEK or,
                The day of week in year of yearly schedule, if yearly_pattern is
                DAY_OF_WEEK. Possible values are: MONDAY, TUESDAY, WEDNESDAY,
                THURSDAY, FRIDAY, SATURDAY, SUNDAY. Defaults to None.
            weekday_offset (ScheduleEnums.WeekdayOffset, optional):
                The weekday offset in month of monthly schedule, if
                monthly_pattern is WEEKDAY. Defaults to None.
            days_of_month (List[str], optional):
                The days of month of monthly schedule, if monthly_pattern is
                DAYS_OF_MONTH. Must be provided as a list of one or more
                stringified digits (from '1' to '31'). Defaults to None.
            monthly_pattern (ScheduleEnums.MonthlyPattern, optional):
                The monthly recurrence pattern of the schedule. Possible values
                are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY, DAYS_OF_MONTH.
                Defaults to None.
            yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                The yearly recurrence pattern of the schedule. Possible values
                are DAY, DAY_OF_WEEK. Defaults to None.
        """
        self.execution.execution_pattern = execution_pattern or self.execution.execution_pattern
        self.recurrence_pattern = recurrence_pattern or self.recurrence_pattern

        if self.execution.execution_pattern == ScheduleEnums.ExecutionPattern.ONCE:
            self.execution = self.Execution.from_dict({
                'execution_pattern': execution_pattern or self.execution.execution_pattern,
                'execution_time': execution_time or self.execution.execution_time
            })
        elif self.execution.execution_pattern == ScheduleEnums.ExecutionPattern.REPEAT:
            self.execution = self.Execution.from_dict({
                'execution_pattern': execution_pattern or self.execution.execution_pattern,
                'start_time': start_time or self.execution.start_time,
                'stop_time': stop_time or self.execution.stop_time,
                'repeat_interval': execution_repeat_interval or self.execution.repeat_interval
            })

        if self.recurrence_pattern == ScheduleEnums.RecurrencePattern.DAILY:
            if hasattr(self, 'daily') and self.daily:
                self.daily = self.Daily.from_dict({
                    'daily_pattern': daily_pattern or self.daily.daily_pattern,
                    'repeat_interval': (repeat_interval
                                        or getattr(self.daily, 'repeat_interval', None))
                })
            else:
                self.daily = self.Daily.from_dict({
                    'daily_pattern': daily_pattern,
                    'repeat_interval': repeat_interval
                })

        if self.recurrence_pattern == ScheduleEnums.RecurrencePattern.WEEKLY:
            if hasattr(self, 'weekly') and self.weekly:
                self.weekly = self.Weekly.from_dict({
                    'repeat_interval': repeat_interval or self.weekly.repeat_interval,
                    'days_of_week': days_of_week or self.weekly.days_of_week
                })
            else:
                self.weekly = self.Weekly.from_dict({
                    'repeat_interval': repeat_interval,
                    'days_of_week': days_of_week
                })

        if self.recurrence_pattern == ScheduleEnums.RecurrencePattern.MONTHLY:
            if hasattr(self, 'monthly') and self.monthly:
                self.monthly = self.Monthly.from_dict({
                    "monthly_pattern": monthly_pattern or self.monthly.monthly_pattern,
                    "repeat_interval": repeat_interval or self.monthly.repeat_interval,
                    "day": day or self.monthly.day,
                    "week_offset": week_offset or getattr(self.monthly, 'week_offset', None),
                    "day_of_week": day_of_week or getattr(self.monthly, 'day_of_week', None),
                    "weekday_offset": (weekday_offset
                                       or getattr(self.monthly, 'weekday_offset', None)),
                    "days_of_month": days_of_month or getattr(self.monthly, 'days_of_month', None)
                })
            else:
                self.monthly = self.Monthly.from_dict({
                    "monthly_pattern": monthly_pattern,
                    "repeat_interval": repeat_interval,
                    "day": day,
                    "week_offset": week_offset,
                    "day_of_week": day_of_week,
                    "weekday_offset": weekday_offset,
                    "days_of_month": days_of_month
                })

        if self.recurrence_pattern == ScheduleEnums.RecurrencePattern.YEARLY:
            if hasattr(self, 'yearly') and self.yearly:
                self.yearly = self.Yearly.from_dict({
                    'yearly_pattern': yearly_pattern or self.yearly.yearly_pattern,
                    'month': month or getattr(self.yearly, 'month', None),
                    'day': day or getattr(self.yearly, 'day', None),
                    'week_offset': week_offset or getattr(self.yearly, 'week_offset', None),
                    'day_of_week': day_of_week or getattr(self.yearly, 'day_of_week', None),
                })
            else:
                self.yearly = self.Yearly.from_dict({
                    'yearly_pattern': yearly_pattern,
                    'month': month,
                    'day': day,
                    'week_offset': week_offset,
                    'day_of_week': day_of_week
                })



from typing import Optional
import json

from mstrio.connection import Connection
from mstrio.server.project import Project
from mstrio.utils.datasources import (alter_conn_list_resp, alter_conn_resp,
                                      alter_instance_list_resp, alter_instance_resp,
                                      alter_patch_req_body)
from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.helper import exception_handler, response_handler, deprecation_warning


@ErrorHandler(err_msg='Error getting available DBMSs.')
def get_available_dbms(connection, error_msg=None):
    """Get information for all available database management systems (DBMSs).

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/dbobjects/dbmss"
    return connection.session.get(url=url)


@ErrorHandler(err_msg='Error getting available database drivers.')
def get_available_db_drivers(connection, error_msg=None):
    """Get information for all available database drivers.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/dbobjects/drivers"
    return connection.session.get(url=url)


def get_datasource_instance(connection, id, error_msg=None):
    """Get information for a specific database source.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/datasources/{id}"
    response = connection.session.get(url=url)
    if not response.ok:
        if error_msg is None:
            error_msg = f"Error getting Datasource Instance with ID: {id}"
        response_handler(response, error_msg)
    response = alter_instance_resp(response)
    return response


@ErrorHandler(err_msg='Error deleting Datasource Instance with ID {id}')
def delete_datasource_instance(connection, id, error_msg=None):
    """Delete a specific database source based on id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 204/400
    """
    url = f"{connection.base_url}/api/datasources/{id}"
    return connection.session.delete(url=url)


def update_datasource_instance(connection, id, body, error_msg=None):
    """Update a specific database source based on id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        body: update operation info
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/datasources/{id}"
    for op_dict in body["operationList"]:
        op_dict = alter_patch_req_body(op_dict, "/datasourceConnection", "/databaseConnectionId")
        op_dict = alter_patch_req_body(op_dict, "/primaryDatasource",
                                       "/databasePrimaryDatasourceId")
        op_dict = alter_patch_req_body(op_dict, "/dataMartDatasource",
                                       "/databaseDataMartDatasourceId")
        alter_patch_req_body(op_dict, "/dbms", "/dbmsId")
    response = connection.session.patch(url=url, json=body)
    if not response.ok:
        if error_msg is None:
            error_msg = f"Error updating Datasource Instance with ID: {id}"
        response_handler(response, error_msg)
    return response


def create_datasource_instance(connection, body, error_msg=None):
    """Create a specific database source.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        body: Datasource info
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 201/400
    """
    url = f"{connection.base_url}/api/datasources"
    response = connection.session.post(url=url, json=body)
    if not response.ok:
        if error_msg is None:
            name = body.get("name", "NA")
            error_msg = f"Error creating Datasource Instance: {name}"
        response_handler(response, error_msg)
    response = alter_instance_resp(response)
    return response


def get_datasource_instances(connection, ids=None, database_type=None, project=None,
                             error_msg=None, application=None):
    """Get information for all database sources.

    Args:
        connection: MicroStrategy REST API connection object
        ids: list of datasources ids
        database_type: list of types (names) of databases
        application: deprecated. Use project instead.
        project: id (str) of a project or instance of an Project class
            to search for the datasource instances in. When provided, both
            `ids` and `database_types` are ignored. By default `None`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    if application:
        deprecation_warning(
            '`application`',
            '`project`',
            '11.3.4.101',  # NOSONAR
            False)
        project = project or application
    project = project.id if isinstance(project, Project) else project
    project_provided = project is not None

    if project_provided:
        url = f"{connection.base_url}/api/projects/{project}/datasources"
        response = connection.session.get(url=url)
    else:
        database_type = None if database_type is None else database_type.join(",")
        ids = None if ids is None else ids.join(",")
        url = f"{connection.base_url}/api/datasources"
        response = connection.session.get(url=url, params={
            'id': ids,
            'database.type': database_type
        })
    if not response.ok:
        res = response.json()
        if project_provided and res.get("message") == "HTTP 404 Not Found":
            # aka project based endpoint not supported
            # try without filtering
            warning_msg = ("get_datasource_instances() warning: filtering by Project "
                           "is not yet supported on this version of the I-Server. "
                           "Returning all values.")
            exception_handler(warning_msg, Warning, 0)
            return get_datasource_instances(connection=connection, ids=ids,
                                            database_type=database_type, error_msg=error_msg)
        if error_msg is None:
            if project_provided \
                    and res.get('code') == "ERR006" \
                    and "not a valid value for Project ID" in res.get('message'):
                error_msg = f"{project} is not a valid Project class instance or ID"
                raise ValueError(error_msg)
            error_msg = "Error getting Datasource Instances"
            if project_provided:
                error_msg += f" within `{project}` Project"
        response_handler(response, error_msg)
    response = alter_instance_list_resp(response)
    return response


def get_datasource_connections(connection, error_msg=None):
    """Get information for all datasource connections.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/datasources/connections"

    response = connection.session.get(url=url)
    if not response.ok:
        if error_msg is None:
            error_msg = "Error getting Datasource Connections"
        response_handler(response, error_msg)
    response = alter_conn_list_resp(response)
    return response


def get_datasource_connection(connection, id, error_msg=None):
    """Get a datasource connection for given id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/datasources/connections/{id}"
    response = connection.session.get(url=url)
    if not response.ok:
        if error_msg is None:
            error_msg = f"Error getting Datasource Connection with ID: {id}"
        response_handler(response, error_msg)
    response = alter_conn_resp(response)
    return response


def update_datasource_connection(connection, id, body, error_msg=None):
    """Update a datasource connection based on id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        body: update operation info
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/datasources/connections/{id}"
    for op_dict in body["operationList"]:
        alter_patch_req_body(op_dict, "/datasourceLogin", "/databaseLoginId")
    response = connection.session.patch(url=url, json=body)
    if not response.ok:
        if error_msg is None:
            error_msg = f"Error updating Datasource Connection with ID: {id}"
        response_handler(response, error_msg)
    return response


@ErrorHandler(err_msg='Error deleting Datasource Connection with ID {id}')
def delete_datasource_connection(connection, id):
    """Delete a datasource connection based on id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 204/400
    """
    url = f"{connection.base_url}/api/datasources/connections/{id}"
    return connection.session.delete(url=url)


def create_datasource_connection(connection, body, error_msg=None):
    """Create a specific database connection.

    Args:
        connection: MicroStrategy REST API connection object
        body: Datasource Connection info
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 201/400
    """
    url = f"{connection.base_url}/api/datasources/connections"
    response = connection.session.post(url=url, json=body)
    if not response.ok:
        if error_msg is None:
            name = body.get("name", "NA")
            error_msg = f"Error creating Datasource connection: {name}"
        response_handler(response, error_msg)
    response = alter_conn_resp(response)
    return response


@ErrorHandler(err_msg='Error testing Datasource connection.')
def test_datasource_connection(connection, body, error_msg=None):
    """Test a datasource connection. Either provide a connection id, or the
    connection parameters within connection object.

    Args:
        connection: MicroStrategy REST API connection object.
        body: Datasource Connection info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 204/400
    """
    url = f"{connection.base_url}/api/datasources/connections/test"
    return connection.session.post(url=url, json=body)


@ErrorHandler(err_msg='Error fetching Datasource mappings.')
def get_datasource_mappings(connection: Connection, default_connection_map: Optional[bool] = False,
                            project_id: Optional[str] = None, error_msg: Optional[str] = None,
                            application_id: Optional[str] = None):
    """Get information for all datasource connection mappings.

    Args:
        connection: MicroStrategy REST API connection object
        default_connection_map (bool, optional): If True will get the default
            connection map for an project. Requires `project_id`
            parameter. Default False.
        project_id: The project_id, required only for default connection
            map.
        application_id: deprecated. Use project_id instead.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    if application_id:
        deprecation_warning(
            '`application_id`',
            '`project_id`',
            '11.3.4.101',  # NOSONAR
            False)
        project_id = project_id or application_id
    url = f"{connection.base_url}/api/datasources/mappings"
    return connection.session.get(
        url=url, params={
            "defaultConnectionMap": default_connection_map,
            "projectId": project_id
        })


# This is a 'fake' get single resource endpoint. Only listing all mappings
# is available on REST API.
# TODO Change to normal get, when it will be available on REST.
@ErrorHandler(err_msg='Error fetching Datasource mapping with ID {id}.')
def get_datasource_mapping(connection: Connection, id=str,
                           default_connection_map: Optional[bool] = False,
                           project_id: Optional[str] = None, error_msg: Optional[str] = None):
    """Get information about specific datasource_mapping.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the mapping
        default_connection_map (bool, optional): If True will get the default
            connection map for an application. Requires `project_id`
            parameter. Default False.
        project_id: The project_id, required only for default connection
            map.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/datasources/mappings"
    response = connection.session.get(
        url=url, params={
            "defaultConnectionMap": default_connection_map,
            "projectId": project_id
        })

    # Faking get single resource endpoint. Only 'list all' available on REST
    if response.ok:
        response_json = response.json()
        if 'mappings' in response_json.keys():
            mappings = [mapping for mapping in response_json['mappings'] if mapping["id"] == id]
            if len(mappings) > 0:
                response_json = mappings[0]
                response_json['ds_connection'] = response_json.pop('connection')
        response.encoding, response._content = 'utf-8', json.dumps(response_json).encode('utf-8')
    return response


@ErrorHandler(err_msg='Error creating Datasource mapping.')
def create_datasource_mapping(connection: Connection, body, error_msg: Optional[str] = None):
    """Create a new datasource mapping.

    Args:
        connection: MicroStrategy REST API connection object
        body: Datasource Connection Map creation info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    url = f"{connection.base_url}/api/datasources/mappings"
    return connection.session.post(url=url, json=body)


@ErrorHandler(err_msg='Error deleting Datasource mapping with ID {id}')
def delete_datasource_mapping(connection: Connection, id: str, error_msg: Optional[str] = None):
    """Delete a datasource mapping based on id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of the mapping meant to be deleted.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    url = f"{connection.base_url}/api/datasources/mappings/{id}"
    return connection.session.delete(url=url)


@ErrorHandler(err_msg="Error getting Datasource logins.")
def get_datasource_logins(connection: Connection, error_msg: Optional[str] = None):
    """Get information for all datasource logins.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/datasources/logins"
    return connection.session.get(url=url)


@ErrorHandler(err_msg='Error creating Datasource login.')
def create_datasource_login(connection: Connection, body, error_msg: Optional[str] = None):
    """Create a new datasource login.

    Args:
        connection: MicroStrategy REST API connection object
        body: Datasource login creation info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    url = f"{connection.base_url}/api/datasources/logins"
    return connection.session.post(url=url, json=body)


@ErrorHandler(err_msg='Error getting Datasource login with ID {id}')
def get_datasource_login(connection: Connection, id: str, error_msg: Optional[str] = None):
    """Get datasource login for a specific id.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the login
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/datasources/logins/{id}"
    return connection.session.get(url=url)


@ErrorHandler(err_msg='Error deleting Datasource login with ID {id}')
def delete_datasource_login(connection: Connection, id: str, error_msg: Optional[str] = None):
    """Delete a datasource login.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the login
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    url = f"{connection.base_url}/api/datasources/logins/{id}"
    return connection.session.delete(url=url)


@ErrorHandler(err_msg='Error updating Datasource login with ID {id}')
def update_datasource_login(connection: Connection, id: str, body,
                            error_msg: Optional[str] = None):
    """Update a datasource login.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the login
        body: Datasource Connection Map creation info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/datasources/logins/{id}"
    return connection.session.patch(url=url, json=body)



from enum import Enum
from typing import List, Optional, TYPE_CHECKING, Union

from mstrio import config
from mstrio.api import datasources, objects
from mstrio.datasources import DatasourceConnection, Dbms
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.helper import get_objects_id
from mstrio.utils.entity import DeleteMixin, CopyMixin, Entity, ObjectTypes

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.server.project import Project


def list_datasource_instances(connection: "Connection", to_dictionary: bool = False,
                              limit: Optional[int] = None, ids: Optional[List[str]] = None,
                              database_types: Optional[List[str]] = None,
                              project: Optional[Union["Project", str]] = None,
                              application: Optional[Union["Project", str]] = None,
                              **filters) -> Union[List["DatasourceInstance"], List[dict]]:
    """Get list of DatasourceInstance objects or dicts. Optionally filter the
    datasource instances by specifying filters.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        to_dictionary: If True returns dict, by default (False) returns
            User objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        ids: list of datasources ids as strings. By default `None`.
        database_types: list of strings representing database types.
            By default `None`. Available values: ['reserved', 'access',
            'altibase', 'amazon_athena', 'amazon_aurora', 'amazon_redshift',
            'arcadia_platform', 'aster', 'big_data_engine', 'cassandra',
            'cirro', 'cloud_element', 'cloud_gateway', 'cloud_gateway_aws_s3',
            'cloud_gateway_azure_adls_2', 'cloud_gateway_google_cloud_storage',
            'composite', 'concur', 'connection_cloud', 'data_direct_cloud',
            'datallegro', 'db2', 'denodo', 'drill', 'dropbox', 'eloqua',
            'enterprise_db', 'ess_base', 'exa_solution', 'excel', 'facebook',
            'gbase_8a', 'generic', 'generic_data_connector', 'google_analytics',
            'google_big_query', 'google_big_query_ff_sql', 'google_drive',
            'hive', 'hive_thrift', 'hubspot', 'impala', 'informatica',
            'informix', 'kognitiowx2', 'kyvos_mdx', 'mapd', 'marketo',
            'mark_logic', 'mem_sql', 'metamatrix', 'microsoft_as', 'mongo_db',
            'my_sql', 'neoview', 'netezza', 'open_access', 'oracle',
            'par_accel', 'par_stream', 'paypal', 'phoenix', 'pig',
            'pivotal_hawq', 'postgre_sql', 'presto', 'red_brick', 'salesforce',
            'sand', 'sap', 'sap_hana', 'sap_hana_mdx', 'search_engine',
            'servicenow', 'shopify', 'snow_flake', 'spark_config', 'spark_sql',
            'splunk', 'sql_server', 'square', 'sybase', 'sybase_iq',
            'sybase_sql_any', 'tandem', 'teradata', 'tm1', 'twitter', 'unknown',
            'url_auth', 'vectorwise', 'vertica', 'xquery']
        project: id (str) of a project or instance of an Project class
            to search for the datasource instances in. When provided, both
            `ids` and `database_types` are ignored. By default `None`.
        application: deprecated. Use project instead.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'datasource_type', table_prefix,
            'odbc_version', 'intermediate_store_db_name',
            'intermediate_store_table_space_name', 'dbms', 'owner', 'acg']

    Examples:
        >>> list_datasource_instances(connection, name='ds_instance_name')
    """
    if application:
        helper.deprecation_warning(
            '`application`',
            '`project`',
            '11.3.4.101',  # NOSONAR
            False)
        project = project or application
    return DatasourceInstance._list_datasource_instances(
        connection=connection,
        ids=ids,
        database_types=database_types,
        to_dictionary=to_dictionary,
        limit=limit,
        project=project,
        **filters,
    )


class DatasourceType(Enum):
    RESERVED = "reserved"
    NORMAL = "normal"
    DATA_IMPORT = "data_import"
    DATA_IMPORT_PRIMARY = "data_import_primary"


class DatasourceInstance(Entity, CopyMixin, DeleteMixin):
    """Object representation of MicroStrategy DataSource Instance object.

    Attributes:
        connection: A MicroStrategy connection object
        id: Datasource Instance ID
        name: Datasource Instance name
        description: Datasource Instance description
        dbms: The database management system (DBMS) object
        database: `Database` object
        datasource_type: DatasourceType Enum (reserved, normal, data_import,
            data_import_primary)
        table_prefix: Table prefix
        odbc_version: Odbc version ENUM
        intermediate_store_db_name: Intermediate store DBName
        intermediate_store_table_space_name: intermediate store table space
            name
        type: Object type
        subtype: Object subtype
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        owner: User object that is the owner
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _OBJECT_TYPE = ObjectTypes.DBROLE
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP, 'dbms': Dbms.from_dict,
        'owner': User.from_dict,
        'datasource_type': DatasourceType,
        'datasource_connection': DatasourceConnection.from_dict
    }
    _API_GETTERS = {
        ('abbreviation', 'type', 'subtype', 'ext_type', 'version', 'owner', 'icon_path',
         'view_media', 'ancestors', 'certified_info', 'acl'): objects.get_object_info,
        ('id', 'name', 'description', 'date_created', 'date_modified', 'datasource_type',
         'table_prefix', 'odbc_version', 'intermediate_store_db_name', 'datasource_connection',
         'database_type', 'database_version', 'primary_datasource', 'data_mart_datasource',
         'intermediate_store_table_space_name', 'dbms', 'acg'): datasources.get_datasource_instance
    }
    _API_PATCH: dict = {
        ('abbreviation'): (objects.update_object, 'partial_put'),
        ("name", "description", "datasource_type", "table_prefix", "intermediate_store_db_name",
         "intermediate_store_table_space_name", "odbc_version", "datasource_connection",
         "primary_datasource", "data_mart_datasource", "dbms"): (
            datasources.update_datasource_instance,
            'patch',
        )
    }
    _PATCH_PATH_TYPES = {
        "name": str,
        "description": str,
        "datasource_type": str,
        "table_prefix": int,
        "intermediate_store_db_name": str,
        "intermediate_store_table_space_name": str,
        "odbc_version": str,
        "datasource_connection": str,
        "primary_datasource": str,
        "data_mart_datasource": str,
        "dbms": str,
    }

    def __init__(self, connection: "Connection", name: Optional[str] = None,
                 id: Optional[str] = None) -> None:
        """Initialize DatasourceInstance object by passing name or id.

        To explore all available DatasourceInstance objects use the
        `list_datasource_instance()` method.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name: exact name of Datasource Instance
            id: ID of Datasource Instance
        """
        if id is None and name is None:
            helper.exception_handler(
                "Please specify either 'id' or 'name' parameter in the constructor.")

        if id is None:
            objects_info = DatasourceInstance._list_datasource_instances(
                connection=connection, name=name, to_dictionary=True)
            if objects_info:
                id, object_info = objects_info[0].pop("id"), objects_info[0]
                super().__init__(connection=connection, object_id=id, **object_info)
            else:
                helper.exception_handler(f"There is no Datasource Instance: '{name}'",
                                         exception_type=ValueError)
        else:
            super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.datasource_type = DatasourceType(
            kwargs["datasource_type"]) if kwargs.get("datasource_type") else None
        self.table_prefix = kwargs.get("table_prefix")
        self.intermediate_store_db_name = kwargs.get("intermediate_store_db_name")
        self.intermediate_store_table_space_name = kwargs.get(
            "intermediate_store_table_space_name")
        self.odbc_version = kwargs.get("odbc_version")
        self.datasource_connection = DatasourceConnection.from_dict(
            kwargs.get("datasource_connection"),
            self.connection) if kwargs.get("datasource_connection") else None
        self.database_type = kwargs.get("database_type")
        self.database_version = kwargs.get("database_version")
        self.primary_datasource = DatasourceInstance.from_dict(
            kwargs.get("primary_datasource"),
            self.connection) if kwargs.get("primary_datasource") else None
        self.data_mart_datasource = DatasourceInstance.from_dict(
            kwargs.get("data_mart_datasource"),
            self.connection) if kwargs.get("data_mart_datasource") else None
        self.dbms = Dbms.from_dict(kwargs.get("dbms"), self.connection)\
            if kwargs.get("dbms") else None

    @classmethod
    def create(
        cls, connection: "Connection", name: str, dbms: Union[Dbms, str],
        description: Optional[str] = None, datasource_type: Optional[Union[str,
                                                                           DatasourceType]] = None,
        table_prefix: Optional[str] = None, odbc_version: Optional[str] = None,
        intermediate_store_db_name: Optional[str] = None,
        intermediate_store_table_space_name: Optional[str] = None,
        datasource_connection: Union[str, DatasourceConnection,
                                     None] = None, database_type: str = None,
        database_version: str = None, primary_datasource: Union[str, "DatasourceInstance",
                                                                None] = None,
        data_mart_datasource: Union[str, "DatasourceInstance", None] = None
    ) -> Optional["DatasourceInstance"]:
        """Create a new DatasourceInstance object on I-Server.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name: Datasource name
            dbms: The database management system (DBMS) object or id
            description: Datasource description
            datasource_type: DatasourceType Enum (reserved, normal, data_import,
                data_import_primary)
            table_prefix: Table prefix
            odbc_version: Odbc version ENUM (version3x, version2x)
            intermediate_store_db_name: Intermediate store DBName
            intermediate_store_table_space_name: intermediate store table space
                name
            datasource_connection: `DatasourceConnection` object or ID
            database_type: Databse type
            database_version: Databse version
            primary_datasource: `DatasourceInstance` object or ID
            data_mart_datasource: `DatasourceInstance` object or ID

        Returns:
            DatasourceInstance object.
        """
        dbms_id = get_objects_id(dbms, Dbms)
        connection_id = get_objects_id(datasource_connection, DatasourceConnection)
        primary_datasource_id = get_objects_id(primary_datasource, cls)
        data_mart_datasource_id = get_objects_id(data_mart_datasource, cls)
        database = {
            "type": database_type,
            "version": database_version,
            "connection": {
                "id": connection_id
            }
        }
        if primary_datasource_id:
            database["primaryDatasource"] = {"id": primary_datasource_id}
        if data_mart_datasource_id:
            database["dataMartDatasource"] = {"id": data_mart_datasource_id}

        body = {
            "name": name,
            "database": database,
            "description": description,
            "datasourceType": helper.get_enum_val(datasource_type, DatasourceType),
            "tablePrefix": table_prefix,
            "odbcVersion": odbc_version,
            "intermediateStoreDbName": intermediate_store_db_name,
            "intermediateStoreTableSpaceName": intermediate_store_table_space_name,
            "dbms": {
                "id": dbms_id
            }
        }
        body = helper.delete_none_values(body)
        response = datasources.create_datasource_instance(connection, body).json()
        if config.verbose:
            print("Successfully created datasource instance named: '{}' with ID: '{}'".format(
                response.get('name'), response.get('id')))
        return cls.from_dict(source=response, connection=connection)

    def alter(self, name: Optional[str] = None, description: Optional[str] = None,
              datasource_type: Optional[Union[str, DatasourceType]] = None,
              table_prefix: Optional[str] = None, odbc_version: Optional[str] = None,
              intermediate_store_db_name: Optional[str] = None,
              intermediate_store_table_space_name: Optional[str] = None, dbms: Union[str, Dbms,
                                                                                     None] = None,
              datasource_connection: Union[str, DatasourceConnection, None] = None,
              primary_datasource: Union[str, "DatasourceInstance", None] = None,
              data_mart_datasource: Union[str, "DatasourceInstance", None] = None) -> None:
        """Alter DatasourceInstance properties.

        Args:
            name: Datasource name
            description: Datasource description
            datasource_type: DatasourceType Enum (reserved, normal, data_import,
                data_import_primary)
            table_prefix: Table prefix
            odbc_version: Odbc version ENUM (version3x, version2x)
            intermediate_store_db_name: Intermediate store DBName
            intermediate_store_table_space_name: intermediate store table space
                name
            dbms: The database management system (DBMS) object or ID
            datasource_connection: `DatasourceConnection` object or ID
            primary_datasource: `DatasourceInstance` object or ID
            data_mart_datasource: `DatasourceInstance` object or ID

        """
        func = self.alter
        args = func.__code__.co_varnames[:func.__code__.co_argcount]
        defaults = func.__defaults__  # type: ignore
        default_dict = dict(zip(args[-len(defaults):], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @classmethod
    def _list_datasource_instances(cls, connection: "Connection",
                                   to_dictionary: Optional[bool] = False,
                                   limit: Optional[int] = None, ids: Optional[list] = None,
                                   database_types: Optional[list] = None,
                                   project: Optional[Union["Project", str]] = None,
                                   **filters) -> Union[List["DatasourceInstance"], List[dict]]:
        objects = helper.fetch_objects(connection=connection,
                                       api=datasources.get_datasource_instances,
                                       dict_unpack_value="datasources", limit=limit,
                                       filters=filters, ids=ids, database_types=database_types,
                                       project=project)
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    def _set_object(self, **kwargs) -> None:
        super(DatasourceInstance, self)._set_object(**kwargs)
        if kwargs.get("primary_datasource"):
            setattr(
                self, 'primary_datasource',
                DatasourceInstance(id=kwargs.get("primary_datasource").get("id"),
                                   name=kwargs.get("primary_datasource").get("name"),
                                   connection=self.connection))
        if kwargs.get("data_mart_datasource"):
            setattr(
                self, 'data_mart_datasource',
                DatasourceInstance(id=kwargs.get("data_mart_datasource").get("id"),
                                   name=kwargs.get("data_mart_datasource").get("name"),
                                   connection=self.connection))



from concurrent.futures.thread import ThreadPoolExecutor
from typing import Callable, Optional

from requests_futures.sessions import FuturesSession

from mstrio.api import monitors
from mstrio.server.node import Node
from mstrio.utils.helper import _prepare_objects, auto_match_args, response_handler


def all_nodes_async(connection, async_api: Callable, filters: dict, error_msg: str,
                    unpack_value: Optional[str] = None, limit: Optional[int] = None, **kwargs):
    """Return list of objects fetched async using wrappers in monitors.py
    """

    node = kwargs.get('node_name')
    if not node:
        nodes_response = monitors.get_node_info(connection).json()
        all_nodes = nodes_response['nodes']
        node_names = [node['name'] for node in all_nodes if node['status'] == 'running']
    else:
        node = node.name if isinstance(node, Node) else node
        if isinstance(node, str):
            node_names = [node]

    if kwargs.get('error_msg'):
        error_msg = kwargs.get('error_msg')

    with FuturesSession(executor=ThreadPoolExecutor(max_workers=8),
                        session=connection.session) as session:
        # Extract parameters of the api wrapper and set them using kwargs
        param_value_dict = auto_match_args(
            async_api, kwargs,
            exclude=['connection', 'limit', 'offset', 'future_session', 'error_msg', 'node_name'])
        futures = (async_api(future_session=session, connection=connection, node_name=n,
                             **param_value_dict) for n in node_names)
        objects = []
        for f in futures:
            response = f.result()
            if not response.ok:
                response_handler(response, error_msg, throw_error=False)
            else:
                obj = _prepare_objects(response.json(), filters, unpack_value)
                objects.extend(obj)
    if limit:
        objects = objects[:limit]
    return objects



from typing import List, Optional, TYPE_CHECKING, Union

from mstrio.api import objects
from mstrio.object_management.search_operations import full_search, SearchDomain
from mstrio.types import ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.acl import ACLMixin
from mstrio.utils.certified_info import CertifiedInfo
from mstrio.utils.entity import CertifyMixin, CopyMixin, DeleteMixin, Entity

if TYPE_CHECKING:
    from mstrio.connection import Connection


def list_objects(connection: "Connection", object_type: ObjectTypes,
                 project_id: Optional[str] = None,
                 domain: Union[SearchDomain, int] = SearchDomain.CONFIGURATION,
                 to_dictionary: bool = False, limit: int = None,
                 **filters) -> Union[List["Object"], List[dict]]:
    """Get list of objects or dicts. Optionally filter the
    objects by specifying filters.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        object_type: Object type. Possible values can
            be found in EnumDSSXMLObjectTypes
        project_id: ID of the project for which the objects are to be listed
        domain: domain where the search will be performed,
            such as Local or Project, possible values
            are defined in EnumDSSXMLSearchDomain
        to_dictionary: If True returns dict, by default (False) returns Objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Examples:
        >>> list_objects(connection, object_type=ObjectTypes.USER)
    """
    return Object._list_objects(
        connection=connection,
        object_type=object_type,
        project_id=project_id,
        domain=domain,
        to_dictionary=to_dictionary,
        limit=limit,
        **filters,
    )


class Object(Entity, ACLMixin, CertifyMixin, CopyMixin, DeleteMixin):
    """Class representing a general type object using attributes common for
     all available objects.

    Attributes:
        name: name of the object
        id: object ID
        description: Description of the object
        abbreviation: Object abbreviation
        version: Object version ID
        ancestors: List of ancestor folders
        type: Object type. Enum
        subtype: Object subtype
        ext_type: Object extended type.
        date_created: Creation time, "yyyy-MM-dd HH:mm:ss" in UTC
        date_modified: Last modification time, "yyyy-MM-dd HH:mm:ss" in UTC
        owner: User object that is the owner
        icon_path: Object icon path
        view_media: View media settings
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
        hidden: Specifies whether the object is hidden
        certified_info: Certification status, time of certification, and
            information about the certifier (currently only for document and
            report)
        project_id: project ID
        target_info: target information, only applicable to Shortcut objects
    """
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'certified_info': CertifiedInfo.from_dict,
        'owner': User.from_dict
    }
    _API_GETTERS = {
        ('id', 'name', 'description', 'date_created', 'date_modified', 'acg', 'abbreviation',
         'type', 'subtype', 'ext_type', 'version', 'owner', 'icon_path', 'view_media', 'ancestors',
         'certified_info', 'acl', 'comments', 'project_id', 'hidden',
         'target_info'): objects.get_object_info
    }

    def __init__(self, connection: "Connection", type: ObjectTypes, id: str):
        """Initialize object by ID.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Identifier of an existing object.
            type (ObjectTypes): object type
        """

        super().__init__(connection=connection, object_id=id, type=type)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._OBJECT_TYPE = ObjectTypes(kwargs.get("type")) if ObjectTypes.contains(
            kwargs.get("type")) else ObjectTypes.NONE

    def __str__(self):
        return f"Object named: '{self.name}' with ID: '{self.id}'"

    def alter(self, name: Optional[str] = None, description: Optional[str] = None,
              abbreviation: Optional[str] = None) -> None:
        """Alter the object properties.

        Args:
            name: object name
            description: object description
            abbreviation: abbreviation
        """
        func = self.alter
        args = func.__code__.co_varnames[:func.__code__.co_argcount]
        defaults = func.__defaults__  # type: ignore
        default_dict = dict(zip(args[-len(defaults):], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @classmethod
    def _list_objects(cls, connection: "Connection", object_type: ObjectTypes,
                      project_id: Optional[str] = None, to_dictionary: bool = False,
                      domain: Union[int, SearchDomain] = SearchDomain.CONFIGURATION,
                      limit: Optional[int] = None, **filters) -> Union[List["Object"], List[dict]]:
        objects = full_search(connection, object_types=object_type,
                              project=project_id, domain=domain, limit=limit, **filters)
        if to_dictionary:
            return objects
        return [cls.from_dict(source=obj, connection=connection) for obj in objects]



# flake8: noqa
from mstrio.project_objects.datasets.cube import list_all_cubes, load_cube
from mstrio.project_objects.datasets.cube_cache import (CubeCache, delete_cube_cache,
                                                        delete_cube_caches, list_cube_caches)
from mstrio.project_objects.datasets.olap_cube import list_olap_cubes, OlapCube
from mstrio.project_objects.datasets.super_cube import list_super_cubes, SuperCube
from mstrio.utils.helper import deprecation_warning

deprecation_warning(
    'mstrio.application_objects.datasets',
    'mstrio.project_objects.datasets',
    '11.3.4.101',  # NOSONAR
    False)



from typing import List, Optional, Union

from pandas import DataFrame

from mstrio.api import documents
from mstrio.project_objects.document import Document
from mstrio.server.environment import Environment
from mstrio.utils import helper

from mstrio.connection import Connection


def list_dossiers(connection: Connection, name: Optional[str] = None, to_dictionary: bool = False,
                  to_dataframe: bool = False, limit: Optional[int] = None,
                  **filters) -> List["Dossier"]:
    """Get all Dossiers stored on the server.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Args:
        connection(object): MicroStrategy connection object returned
            by 'connection.Connection()'
        name: exact name of the document to list
        to_dictionary(bool, optional): if True, return Dossiers as
            list of dicts
        to_dataframe(bool, optional): if True, return Dossiers as
            pandas DataFrame
        limit(int): limit the number of elements returned to a sample of
            dossiers
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
            'owner', 'ext_type', 'view_media', 'certified_info', 'project_id']

    Returns:
            List of dossiers.
    """
    if connection.project_id is None:
        msg = ("Please log into a specific project to load dossiers within it. To load "
               "all dossiers across the whole environment use "
               f"{list_dossiers_across_projects.__name__} function")
        raise ValueError(msg)
    return Dossier._list_all(connection, to_dictionary=to_dictionary, name=name, limit=limit,
                             to_dataframe=to_dataframe, **filters)


def list_dossiers_across_projects(connection: Connection, name: Optional[str] = None,
                                  to_dictionary: bool = False, to_dataframe: bool = False,
                                  limit: Optional[int] = None, **filters) -> List["Dossier"]:
    """Get all Dossiers stored on the server.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Args:
        connection(object): MicroStrategy connection object returned
            by 'connection.Connection()'
        name: exact names of the dossiers to list
        to_dictionary(bool, optional): if True, return Dossiers as
            list of dicts
        to_dataframe(bool, optional): if True, return Dossiers as
            pandas DataFrame
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
            'owner', 'ext_type', 'view_media', 'certified_info', 'project_id']

    Returns:
        List of documents.
    """
    project_id_before = connection.project_id
    env = Environment(connection)
    projects = env.list_projects()
    output = []
    for project in projects:
        connection.select_project(project_id=project.id)
        output.extend(
            Dossier._list_all(connection, to_dictionary=to_dictionary, name=name, limit=limit,
                              to_dataframe=to_dataframe, **filters))
        output = list(set(output))
    connection.select_project(project_id=project_id_before)
    return output


class Dossier(Document):

    @classmethod
    def _list_all(cls, connection: Connection, name: Optional[str] = None,
                  to_dictionary: bool = False, to_dataframe: bool = False,
                  limit: Optional[int] = None,
                  **filters) -> Union[List["Dossier"], List[dict], DataFrame]:
        msg = "Error retrieving documents from the environment."
        if to_dictionary and to_dataframe:
            helper.exception_handler(
                "Please select either to_dictionary=True or to_dataframe=True, but not both.",
                ValueError)
        objects = helper.fetch_objects_async(connection, api=documents.get_dossiers,
                                             async_api=documents.get_dossiers_async,
                                             dict_unpack_value='result', limit=limit,
                                             chunk_size=1000, error_msg=msg, filters=filters,
                                             search_term=name)
        if to_dictionary:
            return objects
        elif to_dataframe:
            return DataFrame(objects)
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]



from datetime import datetime, timezone
from enum import Enum
from typing import List, Optional, Union

from mstrio import config
from mstrio.api import objects, schedules
from mstrio.connection import Connection
from mstrio.distribution_services.schedule import ScheduleEnums, ScheduleTime
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import DeleteMixin, Entity, ObjectTypes
from mstrio.utils.time_helper import DatetimeFormats, map_datetime_to_str, map_str_to_datetime
from mstrio.utils.wip import pause_wip_warnings, resume_wip_warnings

warnings_setting = pause_wip_warnings()

from mstrio.distribution_services.event import Event  # noqa E402

resume_wip_warnings(warnings_setting)


def list_schedules(connection: Connection, to_dictionary: bool = False, limit: int = None,
                   **filters) -> Union[List["Schedule"], List[dict]]:
    """List schedule objects or schedule dictionaries. Optionally filter list.

    Args:
        connection(object): MicroStrategy connection object returned by
            'connection.Connection()'
        to_dictionary(bool, optional): if True, return Schedules as
            list of dicts
        limit(int, optional): maximum number of schedules returned.
        **filters: Available filter parameters:['name':,
                                                'id',
                                                'description',
                                                'schedule_type',
                                                'start_date',
                                                'expired']
    Returns:
        Union[List["Schedule"], List[dict]]: [description]
    """

    objects = helper.fetch_objects(connection=connection, api=schedules.list_schedules,
                                   limit=limit, filters=filters, dict_unpack_value='schedules')

    if to_dictionary:
        return objects
    else:
        return [Schedule.from_dict(source=obj, connection=connection) for obj in objects]


class Schedule(Entity, DeleteMixin):
    """Class representation of MicroStrategy Schedule object.

    Attributes:
        connection: A MicroStrategy connection object
        name: Schedule name
        id: Schedule ID
        description: Schedule description
        schedule_type: Schedule type
        schedule_next_delivery: Schedule next delivery date
        time(ScheduleTime): Details of time-based schedule
        event(Event): Details of event-based schedule
    """

    class ScheduleType(Enum):
        """Class representation of a type of a Microstrategy Schedule."""
        EVENT_BASED = 'event_based'
        TIME_BASED = 'time_based'
        NONE = None

    _OBJECT_TYPE = ObjectTypes.SCHEDULE_TRIGGER

    _API_GETTERS: dict = {
        ('id', 'name', 'description', 'expired', 'schedule_type', 'schedule_next_delivery',
         'start_date', 'stop_date', 'time', 'event'): schedules.get_schedule,
        ('abbreviation', 'type', 'subtype', 'ext_type', 'date_created', 'date_modified', 'version',
         'owner', 'icon_path', 'view_media', 'ancestors', 'certified_info', 'acg',
         'acl'): objects.get_object_info
    }

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'schedule_type': ScheduleType,
        'time': ScheduleTime.from_dict,
        'event': Event.from_dict,
        'schedule_next_delivery': DatetimeFormats.YMDHMS,
        'start_date': DatetimeFormats.DATE,
        'stop_date': DatetimeFormats.DATE,
    }
    _API_PATCH: dict = {
        ('abbreviation'): (objects.update_object, 'partial_put'),
        ('name', 'description', 'schedule_type', 'start_date', 'stop_date', 'time', 'event'):
            (schedules.update_schedule, 'put')
    }

    _PATCH_PATH_TYPES = {
        'name': str,
        'description': str,
        'abbreviation': str,
        'schedule_type': ScheduleType,
        'start_date': datetime,
        'stop_date': datetime,
        'time': ScheduleTime,
        'event': Event,
    }

    def __init__(self, connection: Connection, id: str = None, name: str = None) -> None:
        """Initialize the Schedule object, populates it with I-Server data.
        Specify either `id` or `name`. When `id` is provided (not `None`),
        `name` is omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            id: Schedule ID
            name: Schedule name
        Raises:
            AttributeError: when id is None and name is None.
            ValueError: when there is no schedule with a given name on the
            server.
        """

        if id is None and name is None:
            raise AttributeError(
                "Please specify either 'name' or 'id' parameter in the constructor.")
        if id is None:
            objects_info = list_schedules(connection=connection, name=name, to_dictionary=True)
            if objects_info:
                id, object_info = objects_info[0].pop("id"), objects_info[0]
                super().__init__(connection=connection, object_id=id, **object_info)
            else:
                raise ValueError(f"There is no schedule with the given name: '{name}'")
        else:
            super().__init__(connection, id, name=name)

    def _init_variables(self, **kwargs):
        """
        Add Schedule attributes to _AVAILABLE_ATTRIBUTES map.
        Set attributes inherited from Entity object.
        Set attributes related to schedule details.

        Args:
            **kwargs: dict with information about schedule attributes.
        Returns:
            None
        """
        super()._init_variables(**kwargs)
        self.schedule_type = self.ScheduleType(kwargs.get('schedule_type'))
        self.time = ScheduleTime.from_dict(kwargs.get('time')) if (
            self.schedule_type == self.ScheduleType.TIME_BASED and kwargs.get('time')) else None
        self.event = Event.from_dict(kwargs.get('event'), connection=self._connection) if (
            self.schedule_type == self.ScheduleType.EVENT_BASED and kwargs.get('event')) else None
        self._expired = kwargs.get('expired')
        self._schedule_next_delivery = map_str_to_datetime("schedule_next_delivery",
                                                           kwargs.get("schedule_next_delivery"),
                                                           self._FROM_DICT_MAP)
        self.start_date = map_str_to_datetime("start_date", kwargs.get("start_date"),
                                              self._FROM_DICT_MAP)

    def enable(self, stop_date: Union[str, datetime]) -> bool:
        """Enables schedule and sets stop date

        Args:
            stop_date: stop date provided either as a datetime or
            as a string in yyyy-MM-dd format
        Returns:
            Returns `True` if enabled properly, else `False`.
        """
        self._alter_properties(
            stop_date=map_str_to_datetime('stop_date', stop_date, self._FROM_DICT_MAP))

        if config.verbose and not self.expired:
            print(f'Schedule \'{self.name}\' with ID {self.id} has been enabled, '
                  f'until {self.stop_date}.')
            return True
        elif config.verbose:
            print(f'Schedule \'{self.name}\' with ID {self.id} has NOT been enabled.')
            return False

    def disable(self, stop_date: Union[str, datetime] = None) -> bool:
        """ Disable the schedule. Optional `stop_date` sets the date when
            the schedule should be disabled.

        Args:
            stop_date: stop date provided either as a datetime or
            as a string in yyyy-MM-dd format
        Returns:
            Returns `True` if disabled properly. It does not mean that schedule
            is already expired, as it can take up to one day.
            If operation failed, return `False`.
        """
        stop_date = (datetime.now(timezone.utc) if stop_date is None else map_str_to_datetime(
            'stop_date', stop_date, self._FROM_DICT_MAP))
        self._alter_properties(stop_date=stop_date)

        if config.verbose and self.expired:
            print(f'Schedule \'{self.name}\' with ID {self.id} has been disabled.')
            return True
        elif config.verbose and self.stop_date.date() == stop_date.date():
            print(f'Schedule \'{self.name}\' with ID {self.id} has been set for disabling. '
                  f'Depending on the schedule configuration (`event`, `time` and `stop_date`), '
                  f'it will be disabled by day after \'{self.stop_date.date()}\'.')
            return True
        else:
            print(f'Schedule \'{self.name}\' with ID {self.id} has NOT been disabled.')
            return False

    def list_properties(self):
        """List all properties of the object."""
        attributes = {key: self.__dict__[key] for key in self.__dict__ if not key.startswith('_')}
        attributes = {
            **attributes,
            'expired': self._expired,
            'schedule_next_delivery': self._schedule_next_delivery,
            'id': self.id,
            'type': self.type,
            'subtype': self.subtype,
            'ext_type': self.ext_type,
            'date_created': self.date_created,
            'date_modified': self.date_modified,
            'version': self.version,
            'owner': self.owner,
            'icon_path': self.icon_path,
            'view_media': self.view_media,
            'ancestors': self.ancestors,
            'certified_info': self.certified_info,
            'acg': self.acg,
            'acl': self.acl,
        }

        return {
            key: attributes[key] for key in sorted(attributes, key=helper.sort_object_properties)
        }

    @classmethod
    def create(cls, connection: Connection, name: str, schedule_type: ScheduleType,
               start_date: Union[str, datetime], description: Optional[str] = None,
               stop_date: Optional[Union[str, datetime]] = None, event_id: Optional[str] = None,
               time: Optional[ScheduleTime] = None,
               recurrence_pattern: Optional[ScheduleEnums.RecurrencePattern] = None,
               execution_pattern: Optional[ScheduleEnums.ExecutionPattern] = None,
               execution_time: Optional[str] = None, start_time: Optional[str] = None,
               stop_time: Optional[str] = None, execution_repeat_interval: Optional[int] = None,
               daily_pattern: Optional[ScheduleEnums.DailyPattern] = None,
               repeat_interval: Optional[int] = None,
               days_of_week: Optional[List[ScheduleEnums.DaysOfWeek]] = None,
               day: Optional[int] = None, month: Optional[int] = None,
               week_offset: Optional[ScheduleEnums.WeekOffset] = None,
               day_of_week: Optional[ScheduleEnums.DaysOfWeek] = None,
               weekday_off_set: Optional[str] = None, days_of_month: Optional[List[str]] = None,
               monthly_pattern: Optional[ScheduleEnums.MonthlyPattern] = None,
               yearly_pattern: Optional[ScheduleEnums.YearlyPattern] = None):
        """Create a Schedule using provided parameters as data.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule. Possible values are
                DAILY, WEEKLY, MONTHLY, YEARLY. Defaults to None.
            execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                The execution pattern of the schedule. Possible values are ONCE,
                REPEAT. Defaults to None.
            execution_time (str, optional):
                The execution time of the execution day, if execution_pattern
                is ONCE. Format should be HH:mm:ss. Defaults to None.
            start_time (str, optional):
                The start time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to None.
            stop_time (str, optional):
                The stop time of the execution day, if execution_pattern is
                REPEAT.Format should be HH:mm:ss. Defaults to None.
            execution_repeat_interval (int, optional):
                The repeat interval of minutes of the execution day, if
                execution_pattern is REPEAT. Defaults to None.
            daily_pattern (ScheduleEnums.DailyPattern, optional):
                The daily recurrence pattern of the schedule. Possible values
                are DAY, WEEKDAY. Defaults to None.
            repeat_interval (int, optional):
                The repeat interval of days of daily schedule, if daily_pattern
                is DAY. Defaults to None.
            day (int, optional):
                The day in month of monthly schedule, if monthly_pattern is DAY
                or, The day in month of yearly schedule, if yearly_pattern is
                DAY. Defaults to None.
            month (int, optional):
                The month in year of yearly schedule. Defaults to None.
            week_offset (ScheduleEnums.WeekOffset, optional):
                The week offset in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The week offset in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                SECOND, THIRD, FOURTH, LAST. Defaults to None.
            day_of_week (ScheduleEnums.DaysOfWeek, optional):
                The days of week of weekly schedule or, The day of week in month
                of monthly schedule, if monthly_pattern is DAY_OF_WEEK or, The
                day of week in year of yearly schedule, if yearly_pattern is
                DAY_OF_WEEK. Possible values are: MONDAY, TUESDAY, WEDNESDAY,
                THURSDAY, FRIDAY, SATURDAY, SUNDAY. Defaults to None.
            weekday_offset (ScheduleEnums.WeekdayOffset, optional):
                The weekday offset in month of monthly schedule, if
                monthly_pattern is WEEKDAY. Defaults to None.
            days_of_month (List[str], optional):
                The days of month of monthly schedule, if monthly_pattern is
                DAYS_OF_MONTH. Must be provided as a list of one or more
                stringified digits (from '1' to '31'). Defaults to None.
            monthly_pattern (ScheduleEnums.MonthlyPattern, optional):
                The monthly recurrence pattern of the schedule. Possible values
                are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY, DAYS_OF_MONTH.
                Defaults to None.
            yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                The yearly recurrence pattern of the schedule. Possible values
                are DAY, DAY_OF_WEEK. Defaults to None.
        Returns:
            Schedule object with provided parameters.
        """

        time_kwargs = {
            key: val for key, val in locals().items() if val is not None and key not in [
                'event_id', 'connection', 'description', 'name', 'schedule_type', 'start_date',
                'stop_date', 'time', 'self', 'cls'
            ]
        }
        # Event based or Time based logic
        if schedule_type == cls.ScheduleType.EVENT_BASED:
            execution_details = {'type': 'event', 'content': {'id': event_id}}
        elif schedule_type == cls.ScheduleType.TIME_BASED:
            if time is None:
                time = ScheduleTime.from_details(**time_kwargs)
            execution_details = {'type': 'time', 'content': time.to_dict()}

        # Datetime dates to string format conversion
        start_date = map_datetime_to_str(name='start_date', date=start_date,
                                         string_to_date_map=cls._FROM_DICT_MAP)
        stop_date = map_datetime_to_str(name='stop_date', date=stop_date,
                                        string_to_date_map=cls._FROM_DICT_MAP)

        # Create body and send request
        body = {
            'name': name,
            'description': description,
            'schedule_type': schedule_type.value,
            'start_date': start_date,
            'stop_date': stop_date,
            execution_details['type']: execution_details['content']
        }
        body = helper.delete_none_values(body)
        body = helper.snake_to_camel(body)
        # Response is already unpacked in wrapper
        response = schedules.create_schedule(connection, body)
        response = response.json()
        if config.verbose:
            print(f"Created schedule '{name}' with ID: {response['id']}")
        return Schedule.from_dict(source=response, connection=connection)

    def alter(self, name: str = None, description: str = None,
              start_date: Union[str, datetime] = None, stop_date: Union[str, datetime] = None,
              event: Event = None, event_id: str = None, time: ScheduleTime = None,
              recurrence_pattern: ScheduleEnums.RecurrencePattern = None,
              execution_pattern: ScheduleEnums.ExecutionPattern = None, execution_time: str = None,
              start_time: str = None, stop_time: str = None, execution_repeat_interval: int = None,
              daily_pattern: ScheduleEnums.DailyPattern = None, repeat_interval: int = None,
              days_of_week: List[ScheduleEnums.DaysOfWeek] = None, day: int = None,
              month: int = None, week_offset: ScheduleEnums.WeekOffset = None,
              day_of_week: ScheduleEnums.DaysOfWeek = None, weekday_offset: str = None,
              days_of_month: List[str] = None,
              monthly_pattern: ScheduleEnums.MonthlyPattern = None,
              yearly_pattern: ScheduleEnums.YearlyPattern = None) -> None:
        """Alter Schedule properties.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule. Possible values are
                DAILY, WEEKLY, MONTHLY, YEARLY. Defaults to None.
            execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                The execution pattern of the schedule. Possible values are ONCE,
                REPEAT. Defaults to None.
            execution_time (str, optional):
                The execution time of the execution day, if execution_pattern is
                ONCE. Format should be HH:mm:ss. Defaults to None.
            start_time (str, optional):
                The start time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to None.
            stop_time (str, optional):
                The stop time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to None.
            execution_repeat_interval (int, optional):
                The repeat interval of minutes of the execution day, if
                execution_pattern is REPEAT. Defaults to None.
            daily_pattern (ScheduleEnums.DailyPattern, optional):
                The daily recurrence pattern of the schedule. Possible values
                are DAY, WEEKDAY. Defaults to None.
            repeat_interval (int, optional):
                The repeat interval of days of daily schedule, if daily_pattern
                is DAY. Defaults to None.
            day (int, optional):
                The day in month of monthly schedule, if monthly_pattern is DAY
                or, The day in month of yearly schedule, if yearly_pattern is
                DAY. Defaults to None.
            month (int, optional):
                The month in year of yearly schedule. Defaults to None.
            week_offset (ScheduleEnums.WeekOffset, optional):
                The week offset in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The week offset in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                SECOND, THIRD, FOURTH, LAST. Defaults to None.
            day_of_week (ScheduleEnums.DaysOfWeek, optional):
                The days of week of weekly schedule or,
                The day of week in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The day of week in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are: MONDAY,
                TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY.
                Defaults to None.
            weekday_offset (ScheduleEnums.WeekdayOffset, optional):
                The weekday offset in month of monthly schedule, if
                monthly_pattern is WEEKDAY. Defaults to None.
            days_of_month (List[str], optional):
                The days of month of monthly schedule, if monthly_pattern is
                DAYS_OF_MONTH. Must be provided as a list of one or more
                stringified digits (from '1' to '31'). Defaults to None.
            monthly_pattern (ScheduleEnums.MonthlyPattern, optional):
                The monthly recurrence pattern of the schedule. Possible values
                are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY, DAYS_OF_MONTH.
                Defaults to None.
            yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                The yearly recurrence pattern of the schedule. Possible values
                are DAY, DAY_OF_WEEK. Defaults to None.
        Returns:
            None
        """

        properties = {}

        # Event based or Time based logic
        if time is not None:
            properties['schedule_type'] = self.ScheduleType.TIME_BASED
            properties['time'] = time
        elif event is not None:
            properties['schedule_type'] = self.ScheduleType.EVENT_BASED
            properties['event'] = event
        elif event_id is not None:
            properties['schedule_type'] = self.ScheduleType.EVENT_BASED
            properties['event'] = Event(connection=self._connection, id=event_id)
        elif self.schedule_type == self.ScheduleType.TIME_BASED:
            properties['schedule_type'] = self.ScheduleType.TIME_BASED
            self.time.update_properties(recurrence_pattern, execution_pattern, execution_time,
                                        start_time, stop_time, execution_repeat_interval,
                                        daily_pattern, repeat_interval, days_of_week, day, month,
                                        week_offset, day_of_week, weekday_offset, days_of_month,
                                        monthly_pattern, yearly_pattern)
            properties['time'] = self.time

        if name:
            properties['name'] = name
        if description:
            properties['description'] = description
        if start_date:
            properties['start_date'] = map_str_to_datetime('start_date', start_date,
                                                           self._FROM_DICT_MAP)
        if stop_date:
            properties['stop_date'] = map_str_to_datetime('stop_date', stop_date,
                                                          self._FROM_DICT_MAP)

        self._alter_properties(**properties)

        if self.schedule_type == self.ScheduleType.EVENT_BASED and 'time' in self.__dir__():
            delattr(self, 'time')
        elif self.schedule_type == self.ScheduleType.TIME_BASED and 'event' in self.__dir__():
            delattr(self, 'event')

    def delete(self) -> bool:
        """Delete the schedule.

        Returns:
            bool: True if deletion was successful else False.
        """
        # When the server behavior changes, and deleting last schedule
        # from subscription will no longer delete subscription,
        # move 'force' param to method definition
        force = False
        self._delete_confirm_msg = (
            "This schedule may be part of a subscription. "
            "Deleting such a schedule will remove the subscription as well. "
            "This action cannot be undone. "
            f"Are you sure you want to delete the schedule '{self.name}'"
            f" with ID: {self.id}?[Y/N]: ")
        self._delete_success_msg = (f"Deleted schedule '{self.name}' with ID: {self.id}.")

        return super().delete(force=force)

    @property
    def expired(self):
        """Whether or not the schedule is expired"""
        return self._expired

    @property
    def schedule_next_delivery(self):
        """Next delivery date of time based schedule"""
        return self._schedule_next_delivery
